#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import textwrap
from collections.abc import Callable, Iterable
from pprint import PrettyPrinter
from typing import TYPE_CHECKING, Any, cast, Optional, Union

from elementpath.etree import etree_tostring

from xmlschema.exceptions import XMLSchemaException, XMLSchemaWarning, \
    XMLSchemaValueError, XMLSchemaAttributeError, XMLSchemaTypeError
from xmlschema.aliases import ElementType, NsmapType, SchemaType, SchemaElementType, \
    ModelParticleType
from xmlschema.translation import gettext as _
from xmlschema.resources import XMLResource
from xmlschema.utils.etree import etree_getpath, is_etree_element
from xmlschema.utils.qnames import get_prefixed_qname, local_name

if TYPE_CHECKING:
    from .xsdbase import XsdValidator  # noqa: F401
    from .wildcards import XsdAnyElement  # noqa: F401
    from .groups import XsdGroup  # noqa: F401

ValidatorType = Union['XsdValidator', Callable[[Any], None]]


class XMLSchemaValidatorError(XMLSchemaException):
    """
    Base class for XSD validator errors.

    :param validator: the XSD validator.
    :param message: the error message.
    :param elem: the element that contains the error.
    :param source: the XML resource or the decoded data that contains the error.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    """
    _root: Optional[ElementType] = None
    _path: Optional[str] = None
    _sourceline: Optional[int] = None

    # Optional dump of the execution stack that can be set in collected
    # validator errors for debugging purposes.
    stack_trace: Optional[str] = None

    def __init__(self, validator: ValidatorType,
                 message: str,
                 elem: Optional[ElementType] = None,
                 source: Optional[Any] = None,
                 namespaces: Optional[NsmapType] = None) -> None:
        self.validator = validator
        self._message = message
        self.namespaces = namespaces
        self.source = source
        self.elem = elem

    @property
    def message(self) -> str:
        return self._message

    def __str__(self) -> str:
        chunks: list[str] = ['%s:\n' % self.message.rstrip('.:')]
        if self.elem is not None:
            elem_as_string = etree_tostring(self.elem, self.namespaces, '  ', 20)
            if isinstance(elem_as_string, bytes):
                elem_as_string = elem_as_string.decode('utf-8')
            chunks.append("Schema component:\n\n%s\n" % elem_as_string)

        path = self.path
        if path is not None:
            chunks.append("Path: %s\n" % path)

        if self.schema_url is not None:
            chunks.append("Schema URL: %s\n" % self.schema_url)
            if self.origin_url not in (None, self.schema_url):
                chunks.append("Origin URL: %s\n" % self.origin_url)

        return '\n'.join(chunks) if len(chunks) > 1 else chunks[0][:-2]

    @property
    def msg(self) -> str:
        return self.__str__()

    @msg.setter
    def msg(self, msg: str) -> None:
        return

    def __setattr__(self, name: str, value: Any) -> None:
        if name == 'elem':
            if value is None:
                self._sourceline = None
                self._path = None
                super().__setattr__(name, value)
                return

            if not is_etree_element(value):
                msg = _("{!r} attribute requires an Element, not {!r}.")
                raise XMLSchemaValueError(msg.format(name, value))

            self._sourceline = getattr(value, 'sourceline', self._sourceline)
            self._path = None
            if isinstance(self.source, XMLResource) and self.source.is_lazy():
                # Don't save the element of a lazy resource but set the path
                self._path = etree_getpath(
                    elem=value,
                    root=self.source.root,
                    namespaces=self.namespaces,
                    relative=False,
                    add_position=True
                )
                value = None

        elif name == 'source' and isinstance(value, XMLResource):
            self._path = None

        super().__setattr__(name, value)

    @property
    def sourceline(self) -> Optional[int]:
        """XML element *sourceline* if available (lxml Element)."""
        return getattr(self.elem, 'sourceline', self._sourceline)

    @property
    def root(self) -> Optional[ElementType]:
        """The XML resource root element if *source* is set."""
        if isinstance(self.source, XMLResource):
            return self.source.root
        else:
            return self._root

    @root.setter
    def root(self, value: ElementType) -> None:
        if isinstance(self.source, XMLResource):
            msg = _("can't set the 'root' attribute if an XMLResource source is set")
            raise XMLSchemaAttributeError(msg)
        elif not is_etree_element(value):
            msg = _("{!r} attribute requires an Element, not {!r}.")
            raise XMLSchemaTypeError(msg.format('root', type(value)))
        elif self._root is not value:
            self._root = value
            self._path = None

    @property
    def schema_url(self) -> Optional[str]:
        """The schema URL, if available and the *validator* is an XSD component."""
        url: Optional[str]
        try:
            url = self.validator.schema.source.url  # type: ignore[union-attr]
        except AttributeError:
            return getattr(self.validator, 'url', None)  # it's the schema
        else:
            return url

    @property
    def origin_url(self) -> Optional[str]:
        """The origin schema URL, if available and the *validator* is an XSD component."""
        url: Optional[str]
        try:
            url = self.validator.maps.validator.source.url  # type: ignore[union-attr]
        except AttributeError:
            return None
        else:
            return url

    @property
    def path(self) -> Optional[str]:
        """The XPath of the element, if it's not `None` and the XML resource is set."""
        if self._path is None and self.elem is not None and self.root is not None:
            self._path = etree_getpath(
                elem=self.elem,
                root=self.root,
                namespaces=self.namespaces,
                relative=False,
                add_position=True
            )
        return self._path

    def get_elem_as_string(self, indent: str = '', max_lines: Optional[int] = None) -> str:
        """Returns a string representation of elem attribute."""
        kwargs = {
            'elem': self.elem,
            'namespaces': self.namespaces,
            'indent': indent,
            'max_lines': max_lines
        }
        try:
            return cast(str, etree_tostring(**kwargs))  # type: ignore[arg-type]
        except (ValueError, TypeError):
            return indent + repr(self.elem)


class XMLSchemaCircularityError(XMLSchemaValidatorError):
    """
    Raised when a circularity is found building a global component.
    """
    def __init__(self, name: str, elem: ElementType, schema: SchemaType) -> None:
        msg = _("Circular definition detected for xs:{} {!r}.")
        super().__init__(
            validator=schema,
            message=msg.format(local_name(elem.tag), name),
            elem=elem,
            source=schema.source,
            namespaces=schema.namespaces
        )


class XMLSchemaNotBuiltError(XMLSchemaValidatorError, RuntimeError):
    """
    Raised when there is an improper usage attempt of a not built XSD validator.

    :param validator: the XSD validator.
    :param message: the error message.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    """
    def __init__(self, validator: 'XsdValidator',
                 message: str,
                 namespaces: Optional[NsmapType] = None) -> None:
        if namespaces is None:
            namespaces = getattr(validator, 'namespaces', None)
        super().__init__(
            validator=validator,
            message=message,
            elem=getattr(validator, 'elem', None),
            source=getattr(validator, 'source', None),
            namespaces=namespaces
        )


class XMLSchemaParseError(XMLSchemaValidatorError, SyntaxError):
    """
    Raised when an error is found during the building of an XSD validator.

    :param validator: the XSD validator.
    :param message: the error message.
    :param elem: the element that contains the error.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    """
    def __init__(self, validator: 'XsdValidator', message: str,
                 elem: Optional[ElementType] = None,
                 namespaces: Optional[NsmapType] = None) -> None:
        if namespaces is None:
            namespaces = getattr(validator, 'namespaces', None)

        super().__init__(
            validator=validator,
            message=message,
            elem=elem if elem is not None else getattr(validator, 'elem', None),
            source=getattr(validator, 'source', None),
            namespaces=namespaces
        )


class XMLSchemaModelError(XMLSchemaValidatorError, ValueError):
    """
    Raised when a model error is found during the checking of a model group.

    :param group: the XSD model group.
    :param message: the error message.
    """
    def __init__(self, group: 'XsdGroup', message: str) -> None:
        super().__init__(
            validator=group,
            message=message,
            elem=getattr(group, 'elem', None),
            source=getattr(group, 'source', None),
            namespaces=getattr(group, 'namespaces', None)
        )


class XMLSchemaModelDepthError(XMLSchemaModelError):
    """Raised when recursion depth is exceeded while iterating a model group."""

    @property
    def message(self) -> str:
        return f"{self._message} {self.validator!r}."

    def __init__(self, group: 'XsdGroup') -> None:
        msg = "maximum model recursion depth exceeded while iterating"
        super().__init__(group, message=msg)


class XMLSchemaValidationError(XMLSchemaValidatorError, ValueError):
    """
    Raised when the XML data is not validated with the XSD component or schema.
    It's used by decoding and encoding methods. Encoding validation errors do
    not include XML data element and source, so the error is limited to a message
    containing object representation and a reason.

    :param validator: the XSD validator.
    :param obj: the not validated XML data.
    :param reason: the detailed reason of failed validation.
    :param source: the XML resource that contains the error.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    """
    _message = 'failed validating {} with'

    @property
    def message(self) -> str:
        return f'{self._message} {self.validator!r}.'

    # For compatibility with XMLSchemaChildrenValidationError
    invalid_tag: Optional[str] = None

    @property
    def expected_tags(self) -> list[str]: return []

    @property
    def invalid_child(self) -> Optional[ElementType]: return None

    def __init__(self,
                 validator: ValidatorType,
                 obj: Any,
                 reason: Optional[str] = None,
                 source: Optional[Any] = None,
                 namespaces: Optional[NsmapType] = None) -> None:

        if isinstance(obj, str):
            obj_repr = repr(obj.encode('ascii', 'xmlcharrefreplace').decode('utf-8'))
        else:
            obj_repr = repr(obj)

        if len(obj_repr) > 200:
            obj_repr = f"{type(obj)} instance"

        super().__init__(
            validator=validator,
            message=_(self._message).format(obj_repr),
            elem=obj if is_etree_element(obj) else None,
            source=source,
            namespaces=namespaces,
        )
        self.obj = obj
        self.reason = reason

    def __repr__(self) -> str:
        return '%s(reason=%r)' % (self.__class__.__name__, self.reason)

    def __str__(self) -> str:
        chunks: list[str] = ['%s:\n' % self.message.rstrip('.:')]

        if self.reason is not None:
            chunks.append('Reason: %s\n' % self.reason)

        if hasattr(self.validator, 'tostring'):
            component_as_string = self.validator.tostring('  ', 20)
            chunks.append("Schema component:\n\n%s\n" % component_as_string)

        if is_etree_element(self.elem):
            chunks.append(f"Instance type: {type(self.elem)}\n")
            instance_as_string = self.get_elem_as_string(indent='  ', max_lines=20)
        else:
            chunks.append(f"Instance type: {type(self.obj)}\n")
            instance_as_string = self.get_obj_as_string(indent='  ', max_lines=20)

        if hasattr(self.elem, 'sourceline'):
            line = getattr(self.elem, 'sourceline')
            chunks.append(f"Instance (line {line!r}):\n\n{instance_as_string}\n")
        else:
            chunks.append(f"Instance:\n\n{instance_as_string}\n")

        if self.path is not None:
            chunks.append("Path: %s\n" % self.path)

        return '\n'.join(chunks) if len(chunks) > 1 else chunks[0][:-2]

    def get_obj_as_string(self, indent: str = '', max_lines: Optional[int] = None) -> str:
        """
        Return a string representation of obj attribute, with optional indentation
        and an optional limit on lines.
        """
        if is_etree_element(self.obj):
            return self.get_elem_as_string(indent, max_lines)

        pp = PrettyPrinter(indent=2, depth=6)
        obj_as_string = pp.pformat(self.obj)
        if indent:
            obj_as_string = textwrap.indent(obj_as_string, prefix=indent)

        if max_lines and len(obj_as_string.splitlines()) > max_lines:
            obj_as_string = '\n'.join(obj_as_string.splitlines()[:max_lines - 3])
            obj_as_string += f'\n\n{indent}...\n{indent}...'

        return obj_as_string


class XMLSchemaDecodeError(XMLSchemaValidationError):
    """
    Raised when an XML data string is not decodable to a Python object.

    :param validator: the XSD validator.
    :param obj: the not validated XML data.
    :param decoder: the XML data decoder.
    :param reason: the detailed reason of failed validation.
    :param source: the XML resource that contains the error.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    """
    _message = 'failed decoding {} with'

    def __init__(self, validator: Union['XsdValidator', Callable[[Any], None]],
                 obj: Any,
                 decoder: Any,
                 reason: Optional[str] = None,
                 source: Optional[Any] = None,
                 namespaces: Optional[NsmapType] = None) -> None:
        super().__init__(validator, obj, reason, source, namespaces)
        self.decoder = decoder


class XMLSchemaEncodeError(XMLSchemaValidationError):
    """
    Raised when an object is not encodable to an XML data string.

    :param validator: the XSD validator.
    :param obj: the not validated XML data.
    :param encoder: the XML encoder.
    :param reason: the detailed reason of failed validation.
    :param source: the XML resource that contains the error.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    """
    _message = 'failed encoding {} with'

    def __init__(self, validator: Union['XsdValidator', Callable[[Any], None]],
                 obj: Any,
                 encoder: Any,
                 reason: Optional[str] = None,
                 source: Optional[Any] = None,
                 namespaces: Optional[NsmapType] = None) -> None:
        super().__init__(validator, obj, reason, source, namespaces)
        self.encoder = encoder


class XMLSchemaChildrenValidationError(XMLSchemaValidationError):
    """
    Raised when a child element is not validated.

    :param validator: the XSD validator.
    :param elem: the not validated XML element.
    :param index: the child index.
    :param particle: the model particle that generated the error. Maybe the validator itself.
    :param occurs: the particle occurrences.
    :param expected: the expected element tags/object names.
    :param source: the XML resource that contains the error.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    """
    invalid_tag: Optional[str]
    """The tag of the invalid child element, `None` in case of an incomplete content."""

    def __init__(self, validator: 'XsdValidator',
                 elem: ElementType,
                 index: int,
                 particle: ModelParticleType,
                 occurs: int = 0,
                 expected: Optional[Iterable[SchemaElementType]] = None,
                 source: Optional[Any] = None,
                 namespaces: Optional[NsmapType] = None) -> None:

        self.index = index
        self.particle = particle
        self.occurs = occurs
        self.expected = expected

        if namespaces is None:
            namespaces = getattr(validator, 'namespaces', {})

        if index >= len(elem):
            self.invalid_tag = None
            tag = get_prefixed_qname(elem.tag, namespaces, use_empty=False)
            reason = _("The content of element %r is not complete.") % tag
        else:
            self.invalid_tag = elem[index].tag
            tag = get_prefixed_qname(self.invalid_tag, namespaces, use_empty=False)
            reason = _("Unexpected child with tag %r at position %d.") % (tag, index + 1)

        if occurs and particle.min_occurs > occurs:
            reason += " The particle %r occurs %d times but the minimum is %d." % (
                particle, occurs, particle.min_occurs
            )
        elif particle.max_occurs is not None and particle.max_occurs < occurs:
            reason += " The particle %r occurs %r times but the maximum is %r." % (
                particle, occurs, particle.max_occurs
            )

        expected_tags = self.expected_tags

        if not expected_tags:
            pass
        elif len(expected_tags) > 1:
            reason += _(" Tag (%s) expected.") % ' | '.join(repr(tag) for tag in expected_tags)
        elif expected_tags[0].startswith('from '):
            reason += _(" Tag %s expected.") % expected_tags[0]
        else:
            reason += _(" Tag %r expected.") % expected_tags[0]

        super().__init__(validator, elem, reason, source, namespaces)

    @property
    def expected_tags(self) -> list[str]:
        expected_tags: list[str] = []
        if not self.expected:
            return expected_tags

        for xsd_element in self.expected:
            name = xsd_element.display_name
            if name is not None:
                expected_tags.append(name)
            elif hasattr(xsd_element, 'process_contents'):
                wildcard = cast('XsdAnyElement', xsd_element)
                if wildcard.process_contents == 'strict':
                    tmpl = 'from {!r} namespace/s'
                    items = tuple(wildcard.namespace)
                    if len(wildcard.namespace) == 1:
                        expected_tags.append(tmpl.format(items[0]))
                    else:
                        expected_tags.append(tmpl.format(items))

        return expected_tags

    @property
    def invalid_child(self) -> Optional[ElementType]:
        """
        The invalid child element, if any, `None` otherwise. It's `None` in case of
        incomplete content or if the parent has been cleared during lazy validation.
        """
        try:
            return self.elem[self.index] if self.elem is not None else None
        except IndexError:
            return None  # in case of incomplete content or lazy trees


class XMLSchemaStopValidation(XMLSchemaException):
    """Stops the validation process."""


class XMLSchemaIncludeWarning(XMLSchemaWarning):
    """A schema include fails."""


class XMLSchemaImportWarning(XMLSchemaWarning):
    """A schema namespace import fails."""


class XMLSchemaTypeTableWarning(XMLSchemaWarning):
    """Not equivalent type table found in model."""


class XMLSchemaAssertPathWarning(XMLSchemaWarning):
    """An improper XPath expression found in XSD 1.1 assertion."""



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections.abc import MutableMapping, MutableSequence
from typing import TYPE_CHECKING, Any, Union

from xmlschema.exceptions import XMLSchemaTypeError, XMLSchemaValueError

from .base import ElementData, stackable, XMLSchemaConverter

if TYPE_CHECKING:
    from xmlschema.validators import XsdElement


class UnorderedConverter(XMLSchemaConverter):
    """
    Same as :class:`XMLSchemaConverter` but :meth:`XMLSchemaConverter.element_encode`
    returns a dictionary for the content of the element, that can be used directly
    for unordered encoding mode. In this mode the order of the elements in
    the encoded output is based on the model visitor pattern rather than
    the order in which the elements were added to the input dictionary.
    As the order of the input dictionary is not preserved, character data
    between sibling elements are interleaved between tags.
    """
    __slots__ = ()

    @stackable
    def element_encode(self, obj: Any, xsd_element: 'XsdElement', level: int = 0) -> ElementData:
        """
        Extracts XML decoded data from a data structure for encoding into an ElementTree.

        :param obj: the decoded object.
        :param xsd_element: the `XsdElement` associated to the decoded data structure.
        :param level: the level related to the encoding process (0 means the root).
        :return: an ElementData instance.
        """
        if level or not self.preserve_root:
            element_name = None
        elif not isinstance(obj, MutableMapping):
            raise XMLSchemaTypeError(f"A dictionary expected, got {type(obj)} instead.")
        elif len(obj) != 1:
            raise XMLSchemaValueError("The dictionary must have exactly one element.")
        else:
            element_name, obj = next(iter(obj.items()))

        if not isinstance(obj, MutableMapping):
            if xsd_element.type.simple_type is not None:
                return ElementData(xsd_element.name, obj, None, {}, None)
            elif xsd_element.type.mixed and isinstance(obj, (str, bytes)):
                return ElementData(xsd_element.name, None, [(1, obj)], {}, None)
            else:
                return ElementData(xsd_element.name, None, obj, {}, None)

        text = None
        attributes = {}

        # The unordered encoding mode assumes that the values of this dict will
        # all be lists where each item is the content of a single element. When
        # building content_lu, content which is not a list or lists to be placed
        # into a single element (element has a list content type) must be wrapped
        # in a list to retain that structure. Character data are not wrapped into
        # lists because they are divided from the rest of the content into the
        # unordered mode generator function of the ModelVisitor class.
        content_lu: dict[Union[int, str], Any] = {}

        xmlns = self.set_xmlns_context(obj, level)

        if element_name is None:
            tag = xsd_element.name
        else:
            tag = self.unmap_qname(element_name)
            if not xsd_element.is_matching(tag, self.default_namespace):
                raise XMLSchemaValueError("data tag does not match XSD element name")

        for name, value in obj.items():
            if name == self.text_key:
                text = value
            elif self.cdata_prefix is not None and \
                    name.startswith(self.cdata_prefix) and \
                    (index := name[len(self.cdata_prefix):]).isdigit():
                content_lu[int(index)] = value
            elif self.is_xmlns(name):
                continue
            elif self.attr_prefix and \
                    name.startswith(self.attr_prefix) and \
                    (attr_name := name[len(self.attr_prefix):]):
                ns_name = self.unmap_qname(attr_name, xsd_element.attributes)
                attributes[ns_name] = value
            elif not isinstance(value, MutableSequence) or not value:
                ns_name = self.unmap_qname(name, xmlns=self.get_xmlns_from_data(value))
                content_lu[ns_name] = [value]
            elif isinstance(value[0], (MutableMapping, MutableSequence)):
                ns_name = self.unmap_qname(name, xmlns=self.get_xmlns_from_data(value[0]))
                content_lu[ns_name] = value
            else:
                # `value` is a list but not a list of lists or list of dicts.
                ns_name = self.unmap_qname(name)
                xsd_child = xsd_element.match_child(ns_name)
                if xsd_child is not None:
                    if xsd_child.type and xsd_child.type.is_list():
                        content_lu[ns_name] = [value]
                    else:
                        content_lu[ns_name] = value
                elif self.attr_prefix == '' and ns_name in xsd_element.attributes:
                    attributes[ns_name] = value
                else:
                    content_lu[ns_name] = value

        return ElementData(tag, text, content_lu, attributes, xmlns)



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""Helper functions for QNames and namespaces."""
import re
from collections.abc import Iterable, MutableMapping
from functools import cache

from xmlschema.exceptions import XMLSchemaValueError, XMLSchemaTypeError
from xmlschema.names import XML_NAMESPACE
from xmlschema.aliases import NsmapType


def get_namespace(qname: str) -> str:
    """
    Returns the namespace URI associated with a QName in extended form or a local name.
    If the argument is not conformant to QName format returns the empty string, which
    means no namespace.
    """
    try:
        if qname[0] != '{':
            return ''
        namespace, _ = qname[1:].split('}')
    except (IndexError, ValueError):
        return ''
    except TypeError:
        raise XMLSchemaTypeError("the argument must be a string-like object")
    else:
        return namespace


def get_namespace_ext(qname: str, namespaces: NsmapType | None = None) -> str:
    """
    Returns the namespace URI associated with a QName. If a namespace map is
    provided tries to resolve a prefixed QName and then to extract the namespace.

    :param qname: an extended QName or a local name or a prefixed QName.
    :param namespaces: optional mapping from prefixes to namespace URIs.
    """
    if not namespaces:
        return get_namespace(qname)
    else:
        return get_namespace(get_extended_qname(qname, namespaces))


@cache
def get_qname(uri: str | None, name: str) -> str:
    """
    Returns an expanded QName from URI and local part. If any argument has boolean value
    `False` or if the name is already an expanded QName, returns the *name* argument.

    :param uri: namespace URI
    :param name: local or qualified name
    :return: string or the name argument
    """
    try:
        if name[0] in '{./[' or not uri:
            return name
    except IndexError:
        return ''
    except TypeError:
        raise XMLSchemaTypeError("the 2nd argument must be a string-like object")
    else:
        return f'{{{uri}}}{name}'


@cache
def local_name(qname: str) -> str:
    """
    Return the local part of an expanded QName or a prefixed name. If the name
    is `None` or empty returns the *name* argument.

    :param qname: an expanded QName or a prefixed name or a local name.
    """
    try:
        if qname[0] == '{':
            _namespace, qname = qname.split('}')
        elif ':' in qname:
            _prefix, qname = qname.split(':')
    except IndexError:
        return ''
    except ValueError:
        raise XMLSchemaValueError("the argument 'qname' has an invalid value %r" % qname)
    except TypeError:
        raise XMLSchemaTypeError("the argument 'qname' must be a string-like object")
    else:
        return qname


def get_prefixed_qname(qname: str,
                       namespaces: MutableMapping[str, str] | None,
                       use_empty: bool = True) -> str:
    """
    Get the prefixed form of a QName, using a namespace map.

    :param qname: an extended QName or a local name or a prefixed QName.
    :param namespaces: an optional mapping from prefixes to namespace URIs.
    :param use_empty: if `True` use the empty prefix for mapping.
    """
    if not namespaces or not qname or qname[0] != '{':
        return qname

    namespace = get_namespace(qname)
    prefixes = [x for x in namespaces if namespaces[x] == namespace]

    if not prefixes:
        return qname
    elif prefixes[0]:
        return f"{prefixes[0]}:{qname.split('}', 1)[1]}"
    elif len(prefixes) > 1:
        return f"{prefixes[1]}:{qname.split('}', 1)[1]}"
    elif use_empty:
        return qname.split('}', 1)[1]
    else:
        return qname


def get_extended_qname(qname: str, namespaces: MutableMapping[str, str] | None) -> str:
    """
    Get the extended form of a QName, using a namespace map.
    Local names are mapped to the default namespace.

    :param qname: a prefixed QName or a local name or an extended QName.
    :param namespaces: an optional mapping from prefixes to namespace URIs.
    """
    if not namespaces:
        return qname

    try:
        if qname[0] == '{':
            return qname
    except IndexError:
        return qname

    try:
        prefix, name = qname.split(':', 1)
    except ValueError:
        if not namespaces.get(''):
            return qname
        else:
            return f"{{{namespaces['']}}}{qname}"
    else:
        try:
            uri = namespaces[prefix]
        except KeyError:
            return qname
        else:
            return f'{{{uri}}}{name}' if uri else name


def update_namespaces(namespaces: dict[str, str],
                      xmlns: Iterable[tuple[str, str]],
                      root_declarations: bool = False) -> None:
    """
    Update a namespace map without overwriting existing declarations.
    If a duplicate prefix is encountered in a xmlns declaration, and
    this is mapped to a different namespace, adds the namespace using
    a different generated prefix. The empty prefix '' is used only if
    it's declared at root level to avoid erroneous mapping of local
    names. In other cases it uses the prefix 'default' as substitute.

    :param namespaces: the target namespace map.
    :param xmlns: an iterable containing couples of namespace declarations.
    :param root_declarations: provide `True` if the namespace declarations \
    belong to the root element, `False` otherwise (default).
    """
    for prefix, uri in xmlns:
        if not prefix:
            if not uri:
                continue
            elif '' not in namespaces:
                if root_declarations:
                    namespaces[''] = uri
                    continue
            elif namespaces[''] == uri:
                continue
            prefix = 'default'

        while prefix in namespaces:
            if namespaces[prefix] == uri:
                break
            match = re.search(r'(\d+)$', prefix)
            if match:
                index = int(match.group()) + 1
                prefix = prefix[:match.span()[0]] + str(index)
            else:
                prefix += '0'
        else:
            namespaces[prefix] = uri


def get_namespace_map(namespaces: NsmapType | None) -> dict[str, str]:
    """Returns a new and checked namespace map."""
    if namespaces is None:
        return {}

    namespaces = {k: v for k, v in namespaces.items()}
    if namespaces.get('xml', XML_NAMESPACE) != XML_NAMESPACE:
        msg = f"reserved prefix 'xml' can be used only for {XML_NAMESPACE!r} namespace"
        raise XMLSchemaValueError(msg)

    return namespaces



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from .exceptions import XMLSchemaValidatorError, XMLSchemaParseError, \
    XMLSchemaModelError, XMLSchemaModelDepthError, XMLSchemaValidationError, \
    XMLSchemaDecodeError, XMLSchemaEncodeError, XMLSchemaNotBuiltError, \
    XMLSchemaChildrenValidationError, XMLSchemaStopValidation, \
    XMLSchemaIncludeWarning, XMLSchemaImportWarning, \
    XMLSchemaTypeTableWarning, XMLSchemaAssertPathWarning

from .validation import ValidationContext, DecodeContext, EncodeContext, ValidationMixin
from .xsdbase import XsdValidator, XsdComponent, XsdAnnotation, XsdType
from .particles import ParticleMixin
from .assertions import XsdAssert
from .notations import XsdNotation
from .identities import XsdSelector, XsdFieldSelector, XsdIdentity, XsdKeyref, XsdKey, \
    XsdUnique, Xsd11Keyref, Xsd11Key, Xsd11Unique
from .facets import XsdFacet, XsdWhiteSpaceFacet, XsdLengthFacet, XsdMinLengthFacet, \
    XsdMaxLengthFacet, XsdMinExclusiveFacet, XsdMinInclusiveFacet, XsdMaxExclusiveFacet, \
    XsdMaxInclusiveFacet, XsdFractionDigitsFacet, XsdTotalDigitsFacet, \
    XsdExplicitTimezoneFacet, XsdPatternFacets, XsdEnumerationFacets, XsdAssertionFacet
from .wildcards import XsdAnyElement, Xsd11AnyElement, XsdAnyAttribute, Xsd11AnyAttribute, \
    XsdOpenContent, XsdDefaultOpenContent
from .attributes import XsdAttribute, Xsd11Attribute, XsdAttributeGroup
from .simple_types import XsdSimpleType, XsdAtomic, XsdAtomicBuiltin, \
    XsdAtomicRestriction, Xsd11AtomicRestriction, XsdList, XsdUnion, Xsd11Union
from .complex_types import XsdComplexType, Xsd11ComplexType
from .models import ModelVisitor
from .groups import XsdGroup, Xsd11Group
from .elements import XsdElement, Xsd11Element, XsdAlternative

from .builders import XsdBuilders, GlobalMaps
from .xsd_globals import XsdGlobals
from .schemas import XMLSchemaMeta, XMLSchemaBase, XMLSchema, XMLSchema10, XMLSchema11


__all__ = [
    'ValidationContext', 'DecodeContext', 'EncodeContext', 'ValidationMixin',
    'XMLSchemaValidatorError', 'XMLSchemaParseError', 'XMLSchemaModelError',
    'XMLSchemaModelDepthError', 'XMLSchemaValidationError', 'XMLSchemaDecodeError',
    'XMLSchemaEncodeError', 'XMLSchemaNotBuiltError', 'XMLSchemaChildrenValidationError',
    'XMLSchemaStopValidation', 'XMLSchemaIncludeWarning', 'XMLSchemaImportWarning',
    'XMLSchemaTypeTableWarning', 'XMLSchemaAssertPathWarning',
    'XsdValidator', 'XsdComponent', 'XsdAnnotation', 'XsdType',
    'ParticleMixin', 'XsdAssert', 'XsdNotation', 'XsdSelector', 'XsdFieldSelector',
    'XsdIdentity', 'XsdKeyref', 'XsdKey', 'XsdUnique', 'Xsd11Keyref', 'Xsd11Key',
    'Xsd11Unique', 'XsdFacet', 'XsdWhiteSpaceFacet', 'XsdLengthFacet', 'XsdMinLengthFacet',
    'XsdMaxLengthFacet', 'XsdMinExclusiveFacet', 'XsdMinInclusiveFacet',
    'XsdMaxExclusiveFacet', 'XsdMaxInclusiveFacet', 'XsdFractionDigitsFacet',
    'XsdTotalDigitsFacet', 'XsdExplicitTimezoneFacet', 'XsdPatternFacets',
    'XsdEnumerationFacets', 'XsdAssertionFacet', 'XsdAnyElement', 'Xsd11AnyElement',
    'XsdAnyAttribute', 'Xsd11AnyAttribute', 'XsdOpenContent', 'XsdDefaultOpenContent',
    'XsdAttribute', 'Xsd11Attribute', 'XsdAttributeGroup', 'XsdSimpleType', 'XsdAtomic',
    'XsdAtomicBuiltin', 'XsdAtomicRestriction', 'Xsd11AtomicRestriction', 'XsdList',
    'XsdUnion', 'Xsd11Union', 'XsdComplexType', 'Xsd11ComplexType', 'ModelVisitor',
    'XsdGroup', 'Xsd11Group', 'XsdElement', 'Xsd11Element', 'XsdAlternative',
    'XsdBuilders', 'GlobalMaps', 'XsdGlobals',
    'XMLSchemaMeta', 'XMLSchemaBase', 'XMLSchema', 'XMLSchema10', 'XMLSchema11',
]



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from threading import Lock
from collections.abc import Callable
from functools import lru_cache, wraps
from typing import Any, Generic, overload, TypeVar, TYPE_CHECKING, Union, cast

from xmlschema.aliases import SchemaType
from xmlschema.exceptions import XMLSchemaAttributeError, XMLSchemaTypeError, XMLSchemaValueError

if TYPE_CHECKING:
    from xmlschema.validators.xsdbase import XsdComponent  # noqa


T = TypeVar('T', bound=Union[SchemaType, 'XsdComponent'])
RT = TypeVar('RT', covariant=True)


_cached_functions: dict[Callable[..., Any], tuple[int | None, bool]] = {}


class SchemaCache:

    __slots__ = ('_enabled', '_functions', '_caches', '_lock')

    def __init__(self, enabled: bool = True) -> None:
        self._enabled = enabled
        self._functions = _cached_functions.copy()
        self._caches: dict[Callable[..., Any], Any] = {}
        self._lock = Lock()
        self._create_caches()

    def __call__(self, func: Callable[..., RT], *args: Any, **kwargs: Any) -> RT:
        try:
            return cast(RT, self._caches[func](*args, **kwargs))
        except KeyError:
            with self._lock:
                return cast(RT, self._caches[func](*args, **kwargs))

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        if value is not self._enabled:
            self._enabled = value
            with self._lock:
                self._create_caches()

    def _create_caches(self) -> None:
        if self._enabled:
            for func, (maxsize, typed) in self._functions.items():
                self._caches[func] = lru_cache(maxsize, typed)(func)
        else:
            # Caching is not enables: fallback to registered functions
            for func in self._functions:
                self._caches[func] = func

    def register(self, func: Callable[..., RT],
                 maxsize: int | None = None,
                 typed: bool = True) -> None:
        if not callable(func):
            raise XMLSchemaTypeError(f"{func!r} is not callable")

        with self._lock:
            if func in self._functions:
                raise XMLSchemaValueError(f"{func!r} is already cached by {self!r}")
            self._functions[func] = maxsize, typed
            self._caches[func] = lru_cache(maxsize, typed)(func) if self._enabled else func

    def clear(self) -> None:
        with self._lock:
            if self._enabled:
                for cache in self._caches.values():
                    cache.cache_clear()


def schema_lru_cache(maxsize: int | None = None, typed: bool = False) -> Callable[..., RT]:
    """
    Cache an XSD validator method using an LRU cache stored in XSD global maps cache.
    """
    def wrapper_cached(func: Callable[..., RT]) -> Callable[..., RT]:
        _cached_functions[func] = maxsize, typed

        @wraps(func)
        def wrapped_func(validator: T, *args: Any, **kwargs: Any) -> RT:
            return validator.maps.cache(func, validator, *args, **kwargs)
        return wrapped_func

    return cast(Callable[..., RT], wrapper_cached)


def schema_cache(func: Callable[..., RT]) -> Callable[..., RT]:
    """
    Cache an XSD validator method using an unlimited cache stored in XSD global maps cache.
    """
    _cached_functions[func] = None, False

    @wraps(func)
    def wrapped_func(validator: T, *args: Any, **kwargs: Any) -> RT:
        return validator.maps.cache(func, validator, *args, **kwargs)
    return wrapped_func


# noinspection PyPep8Naming
class schema_cached_property(Generic[T, RT]):
    """
    A property that caches the value on the :class:`SchemaCache` of the global maps.
    """
    __slots__ = ('func', '_name', '__dict__')

    def __init__(self, func: Callable[[T], RT]) -> None:
        _cached_functions[func] = None, False
        self.func = func
        self.__doc__ = func.__doc__
        if hasattr(func, '__module__'):
            self.__module__ = func.__module__

    def __set_name__(self, owner: type[T], name: str) -> None:
        if not hasattr(owner, 'built'):
            raise XMLSchemaTypeError("{!r} is not an XSD validator".format(owner))
        if name == 'built':
            raise XMLSchemaAttributeError("can't apply to 'built' property")
        self._name = name

    @overload
    def __get__(self, validator: None, owner: type[T]) -> Callable[[T], RT]:
        ...

    @overload
    def __get__(self, validator: T, owner: type[T]) -> RT:
        ...

    def __get__(self, validator: T | None, owner: type[T]) -> Callable[[T], RT] | RT:
        if validator is None:
            return self.func
        elif not validator.maps.built:
            return self.func(validator)
        return validator.maps.cache(self.func, validator)



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from xml.etree.ElementTree import ParseError


class XMLSchemaException(Exception):
    """The base exception that let you catch all the errors generated by the library."""


class XMLSchemaOSError(XMLSchemaException, OSError):
    pass


class XMLSchemaAttributeError(XMLSchemaException, AttributeError):
    pass


class XMLSchemaTypeError(XMLSchemaException, TypeError):
    pass


class XMLSchemaValueError(XMLSchemaException, ValueError):
    pass


class XMLSchemaKeyError(XMLSchemaException, KeyError):
    pass


class XMLSchemaRuntimeError(XMLSchemaException, RuntimeError):
    pass


class XMLSchemaNamespaceError(XMLSchemaException, RuntimeError):
    """Raised when a wrong runtime condition is found with a namespace."""


class XMLSchemaWarning(Warning):
    """Base warning class for the XMLSchema package."""


class XMLResourceError(XMLSchemaException):
    """
    A generic error on an XML resource that catches all the errors generated
    by an XML resource/loader instance on accessing XML data.
    """


class XMLResourceParseError(XMLResourceError, ParseError):
    """Raised when an error is found parsing an XML resource."""


class XMLResourceExceeded(XMLResourceError):
    """Raised when the parsing of an XML resource exceeds a safety limit."""


class XMLResourceBlocked(XMLResourceError):
    """Raised when an XML resource access is blocked by security settings."""


class XMLResourceForbidden(XMLResourceError):
    """Raised when the parsing of an XML resource is forbidden for safety reasons."""


class XMLResourceOSError(XMLResourceError, OSError):
    """
    Raised when an XMLResource calls a system function that returns a system-related error,
    including I/O operations.
    """



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from typing import Any, Union, cast


from xmlschema.exceptions import XMLSchemaTypeError
from xmlschema.translation import gettext as _
from xmlschema.arguments import Option
from xmlschema.utils.misc import is_subclass

from .base import ElementData, XMLSchemaConverter
from .unordered import UnorderedConverter
from .parker import ParkerConverter
from .badgerfish import BadgerFishConverter
from .gdata import GDataConverter
from .abdera import AbderaConverter
from .jsonml import JsonMLConverter
from .columnar import ColumnarConverter

__all__ = ['XMLSchemaConverter', 'UnorderedConverter', 'ParkerConverter',
           'BadgerFishConverter', 'AbderaConverter', 'JsonMLConverter',
           'ColumnarConverter', 'ElementData', 'GDataConverter',
           'ConverterType', 'ConverterOption']


ConverterType = Union[type[XMLSchemaConverter], XMLSchemaConverter]


class ConverterOption(Option[ConverterType | None]):
    def validated_value(self, value: Any) -> ConverterType | None:
        if value is None or isinstance(value, XMLSchemaConverter) \
                or is_subclass(value, XMLSchemaConverter):
            return cast(ConverterType, value)
        msg = _("invalid type {!r} for {}, must be a {!r} instance/subclass or None")
        raise XMLSchemaTypeError(msg.format(type(value), self, XMLSchemaConverter))



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
This module contains definitions and a factory function for XSD builtin datatypes.

Only atomic builtins are created, the list builtins types ('NMTOKENS', 'ENTITIES', 'IDREFS')
are created using the XSD 1.0 meta-schema or with and additional base schema for XSD 1.1.
"""
from decimal import Decimal
from elementpath import datatypes
from typing import Any
from xml.etree.ElementTree import Element

import xmlschema.names as nm

from .helpers import decimal_validator, qname_validator, byte_validator, \
    short_validator, int_validator, long_validator, unsigned_byte_validator, \
    unsigned_short_validator, unsigned_int_validator, unsigned_long_validator, \
    negative_int_validator, positive_int_validator, non_positive_int_validator, \
    non_negative_int_validator, hex_binary_validator, base64_binary_validator, \
    error_type_validator, boolean_to_python, python_to_boolean, python_to_float, \
    python_to_int

#
# Admitted facets sets for XSD atomic types
STRING_FACETS = {
    nm.XSD_LENGTH, nm.XSD_MIN_LENGTH, nm.XSD_MAX_LENGTH, nm.XSD_PATTERN,
    nm.XSD_ENUMERATION, nm.XSD_WHITE_SPACE, nm.XSD_ASSERTION
}

BOOLEAN_FACETS = {nm.XSD_PATTERN, nm.XSD_WHITE_SPACE, nm.XSD_ASSERTION}

FLOAT_FACETS = {
    nm.XSD_PATTERN, nm.XSD_ENUMERATION, nm.XSD_WHITE_SPACE, nm.XSD_MAX_INCLUSIVE,
    nm.XSD_MAX_EXCLUSIVE, nm.XSD_MIN_INCLUSIVE, nm.XSD_MIN_EXCLUSIVE, nm.XSD_ASSERTION
}

DECIMAL_FACETS = {
    nm.XSD_TOTAL_DIGITS, nm.XSD_FRACTION_DIGITS, nm.XSD_PATTERN, nm.XSD_ENUMERATION,
    nm.XSD_WHITE_SPACE, nm.XSD_MAX_INCLUSIVE, nm.XSD_MAX_EXCLUSIVE, nm.XSD_MIN_INCLUSIVE,
    nm.XSD_MIN_EXCLUSIVE, nm.XSD_ASSERTION
}

DATETIME_FACETS = {
    nm.XSD_PATTERN, nm.XSD_ENUMERATION, nm.XSD_WHITE_SPACE,
    nm.XSD_MAX_INCLUSIVE, nm.XSD_MAX_EXCLUSIVE, nm.XSD_MIN_INCLUSIVE,
    nm.XSD_MIN_EXCLUSIVE, nm.XSD_ASSERTION, nm.XSD_EXPLICIT_TIMEZONE
}

#
# Element facets instances for builtin types.
PRESERVE_WHITE_SPACE_ELEMENT = Element(nm.XSD_WHITE_SPACE, value='preserve')
COLLAPSE_WHITE_SPACE_ELEMENT = Element(nm.XSD_WHITE_SPACE, value='collapse')
REPLACE_WHITE_SPACE_ELEMENT = Element(nm.XSD_WHITE_SPACE, value='replace')
XSD10_FLOAT_PATTERN_ELEMENT = Element(
    nm.XSD_PATTERN,
    value=r"(\+|-)?([0-9]+(\.[0-9]*)?|\.[0-9]+)([Ee](\+|-)?[0-9]+)?|INF|-INF|NaN"
)
XSD11_FLOAT_PATTERN_ELEMENT = Element(
    nm.XSD_PATTERN,
    value=r"(\+|-)?([0-9]+(\.[0-9]*)?|\.[0-9]+)([Ee](\+|-)?[0-9]+)?|(\+|-)?INF|NaN"
)


XSD_COMMON_BUILTIN_TYPES: tuple[dict[str, Any], ...] = (
    # ***********************
    # *** Primitive types ***
    # ***********************

    # --- String Types ---
    {
        'name': nm.XSD_STRING,
        'datatype': datatypes.StringProxy,
        'python_type': str,
        'admitted_facets': STRING_FACETS,
        'facets': [PRESERVE_WHITE_SPACE_ELEMENT],
    },  # character string

    # --- Numerical Types ---
    {
        'name': nm.XSD_DECIMAL,
        'datatype': datatypes.DecimalProxy,
        'python_type': (Decimal, int, float),
        'admitted_facets': DECIMAL_FACETS,
        'to_python': datatypes.DecimalProxy,
        'facets': [decimal_validator, COLLAPSE_WHITE_SPACE_ELEMENT],
    },  # decimal number

    # --- Dates and Times (not year related) ---
    {
        'name': nm.XSD_GDAY,
        'datatype': datatypes.GregorianDay,
        'python_type': datatypes.GregorianDay,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.GregorianDay.fromstring,
    },  # DD
    {
        'name': nm.XSD_GMONTH,
        'datatype': datatypes.GregorianMonth,
        'python_type': datatypes.GregorianMonth,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.GregorianMonth.fromstring,
    },  # MM
    {
        'name': nm.XSD_GMONTH_DAY,
        'datatype': datatypes.GregorianMonthDay,
        'python_type': datatypes.GregorianMonthDay,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.GregorianMonthDay.fromstring,
    },  # MM-DD
    {
        'name': nm.XSD_TIME,
        'datatype': datatypes.Time,
        'python_type': datatypes.Time,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.Time.fromstring,
    },  # hh:mm:ss
    {
        'name': nm.XSD_DURATION,
        'datatype': datatypes.Duration,
        'python_type': datatypes.Duration,
        'admitted_facets': FLOAT_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.Duration.fromstring,
    },  # PnYnMnDTnHnMnS

    # Other primitive types
    {
        'name': nm.XSD_QNAME,
        'datatype': datatypes.QName,
        'python_type': (str, datatypes.QName),
        'admitted_facets': STRING_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT, qname_validator],
    },  # prf:name (the prefix needs to be qualified with an in-scope namespace)
    {
        'name': nm.XSD_NOTATION_TYPE,
        'datatype': datatypes.Notation,
        'python_type': (str, datatypes.Notation),
        'admitted_facets': STRING_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
    },  # type for NOTATION attributes: QNames of xs:notation declarations as value space.
    {
        'name': nm.XSD_ANY_URI,
        'datatype': datatypes.AnyURI,
        'python_type': (str, datatypes.AnyURI),
        'admitted_facets': STRING_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
    },  # absolute or relative uri (RFC 2396)
    {
        'name': nm.XSD_BOOLEAN,
        'datatype': datatypes.BooleanProxy,
        'python_type': bool,
        'admitted_facets': BOOLEAN_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': boolean_to_python,
        'from_python': python_to_boolean,
    },  # true/false or 1/0
    {
        'name': nm.XSD_BASE64_BINARY,
        'datatype': datatypes.Base64Binary,
        'python_type': (datatypes.Base64Binary, str, bytes),
        'admitted_facets': STRING_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT, base64_binary_validator],
    },  # base64 encoded binary value
    {
        'name': nm.XSD_HEX_BINARY,
        'datatype': datatypes.HexBinary,
        'python_type': (datatypes.HexBinary, str, bytes),
        'admitted_facets': STRING_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT, hex_binary_validator],
    },   # hexadecimal encoded binary value

    # *********************
    # *** Derived types ***
    # *********************

    # --- String Types ---
    {
        'name': nm.XSD_NORMALIZED_STRING,
        'datatype': datatypes.NormalizedString,
        'python_type': str,
        'base_type': nm.XSD_STRING,
        'facets': [REPLACE_WHITE_SPACE_ELEMENT],
    },  # line breaks are normalized
    {
        'name': nm.XSD_TOKEN,
        'datatype': datatypes.XsdToken,
        'python_type': str,
        'base_type': nm.XSD_NORMALIZED_STRING,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
    },  # whitespaces are normalized
    {
        'name': nm.XSD_LANGUAGE,
        'datatype': datatypes.Language,
        'python_type': str,
        'base_type': nm.XSD_TOKEN,
        'facets': [Element(nm.XSD_PATTERN, value=r"[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*")]
    },  # language codes
    {
        'name': nm.XSD_NAME,
        'datatype': datatypes.Name,
        'python_type': str,
        'base_type': nm.XSD_TOKEN,
        'facets': [Element(nm.XSD_PATTERN, value=r"\i\c*")]
    },  # not starting with a digit
    {
        'name': nm.XSD_NCNAME,
        'datatype': datatypes.NCName,
        'python_type': str,
        'base_type': nm.XSD_NAME,
        'facets': [Element(nm.XSD_PATTERN, value=r"[\i-[:]][\c-[:]]*")]
    },  # cannot contain colons
    {
        'name': nm.XSD_ID,
        'datatype': datatypes.Id,
        'python_type': str,
        'base_type': nm.XSD_NCNAME
    },  # unique identification in document (attribute only)
    {
        'name': nm.XSD_IDREF,
        'datatype': datatypes.Idref,
        'python_type': str,
        'base_type': nm.XSD_NCNAME
    },  # reference to ID field in document (attribute only)
    {
        'name': nm.XSD_ENTITY,
        'datatype': datatypes.Entity,
        'python_type': str,
        'base_type': nm.XSD_NCNAME
    },  # reference to entity (attribute only)
    {
        'name': nm.XSD_NMTOKEN,
        'datatype': datatypes.NMToken,
        'python_type': str,
        'base_type': nm.XSD_TOKEN,
        'facets': [Element(nm.XSD_PATTERN, value=r"\c+")]
    },  # should not contain whitespace (attribute only)

    # --- Numerical derived types ---
    {
        'name': nm.XSD_INTEGER,
        'datatype': datatypes.Integer,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_DECIMAL
    },  # any integer value
    {
        'name': nm.XSD_LONG,
        'datatype': datatypes.Long,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_INTEGER,
        'facets': [long_validator,
                   Element(nm.XSD_MIN_INCLUSIVE, value='-9223372036854775808'),
                   Element(nm.XSD_MAX_INCLUSIVE, value='9223372036854775807')]
    },  # signed 128 bit value
    {
        'name': nm.XSD_INT,
        'datatype': datatypes.Int,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_LONG,
        'facets': [int_validator,
                   Element(nm.XSD_MIN_INCLUSIVE, value='-2147483648'),
                   Element(nm.XSD_MAX_INCLUSIVE, value='2147483647')]
    },  # signed 64 bit value
    {
        'name': nm.XSD_SHORT,
        'datatype': datatypes.Short,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_INT,
        'facets': [short_validator,
                   Element(nm.XSD_MIN_INCLUSIVE, value='-32768'),
                   Element(nm.XSD_MAX_INCLUSIVE, value='32767')]
    },  # signed 32 bit value
    {
        'name': nm.XSD_BYTE,
        'datatype': datatypes.Byte,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_SHORT,
        'facets': [byte_validator,
                   Element(nm.XSD_MIN_INCLUSIVE, value='-128'),
                   Element(nm.XSD_MAX_INCLUSIVE, value='127')]
    },  # signed 8 bit value
    {
        'name': nm.XSD_NON_NEGATIVE_INTEGER,
        'datatype': datatypes.NonNegativeInteger,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_INTEGER,
        'facets': [non_negative_int_validator, Element(nm.XSD_MIN_INCLUSIVE, value='0')]
    },  # only zero and more value allowed [>= 0]
    {
        'name': nm.XSD_POSITIVE_INTEGER,
        'datatype': datatypes.PositiveInteger,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_NON_NEGATIVE_INTEGER,
        'facets': [positive_int_validator, Element(nm.XSD_MIN_INCLUSIVE, value='1')]
    },  # only positive value allowed [> 0]
    {
        'name': nm.XSD_UNSIGNED_LONG,
        'datatype': datatypes.UnsignedLong,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_NON_NEGATIVE_INTEGER,
        'facets': [unsigned_long_validator,
                   Element(nm.XSD_MAX_INCLUSIVE, value='18446744073709551615')]
    },  # unsigned 128 bit value
    {
        'name': nm.XSD_UNSIGNED_INT,
        'datatype': datatypes.UnsignedInt,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_UNSIGNED_LONG,
        'facets': [unsigned_int_validator, Element(nm.XSD_MAX_INCLUSIVE, value='4294967295')]
    },  # unsigned 64 bit value
    {
        'name': nm.XSD_UNSIGNED_SHORT,
        'datatype': datatypes.UnsignedShort,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_UNSIGNED_INT,
        'facets': [unsigned_short_validator, Element(nm.XSD_MAX_INCLUSIVE, value='65535')]
    },  # unsigned 32 bit value
    {
        'name': nm.XSD_UNSIGNED_BYTE,
        'datatype': datatypes.UnsignedByte,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_UNSIGNED_SHORT,
        'facets': [unsigned_byte_validator, Element(nm.XSD_MAX_INCLUSIVE, value='255')]
    },  # unsigned 8 bit value
    {
        'name': nm.XSD_NON_POSITIVE_INTEGER,
        'datatype': datatypes.NonPositiveInteger,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_INTEGER,
        'facets': [non_positive_int_validator, Element(nm.XSD_MAX_INCLUSIVE, value='0')]
    },  # only zero and smaller value allowed [<= 0]
    {
        'name': nm.XSD_NEGATIVE_INTEGER,
        'datatype': datatypes.NegativeInteger,
        'python_type': int,
        'from_python': python_to_int,
        'base_type': nm.XSD_NON_POSITIVE_INTEGER,
        'facets': [negative_int_validator, Element(nm.XSD_MAX_INCLUSIVE, value='-1')]
    },  # only negative value allowed [< 0]
)

XSD_10_BUILTIN_TYPES: tuple[dict[str, Any], ...] = XSD_COMMON_BUILTIN_TYPES + (
    {
        'name': nm.XSD_DOUBLE,
        'datatype': datatypes.DoubleProxy10,
        'python_type': float,
        'admitted_facets': FLOAT_FACETS,
        'facets': [XSD10_FLOAT_PATTERN_ELEMENT, COLLAPSE_WHITE_SPACE_ELEMENT],
        'from_python': python_to_float,
    },  # 64 bit floating point
    {
        'name': nm.XSD_FLOAT,
        'datatype': datatypes.Float10,
        'python_type': float,
        'admitted_facets': FLOAT_FACETS,
        'facets': [XSD10_FLOAT_PATTERN_ELEMENT, COLLAPSE_WHITE_SPACE_ELEMENT],
        'from_python': python_to_float,
    },  # 32 bit floating point

    # --- Year related primitive types (year 0 not allowed) ---
    {
        'name': nm.XSD_DATETIME,
        'datatype': datatypes.DateTime10,
        'python_type': datatypes.DateTime10,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.DateTime10.fromstring,
    },  # [-][Y*]YYYY-MM-DD[Thh:mm:ss]
    {
        'name': nm.XSD_DATE,
        'datatype': datatypes.Date10,
        'python_type': datatypes.Date10,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.Date10.fromstring,
    },  # [-][Y*]YYYY-MM-DD
    {
        'name': nm.XSD_GYEAR,
        'datatype': datatypes.GregorianYear10,
        'python_type': datatypes.GregorianYear10,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.GregorianYear10.fromstring,
    },  # [-][Y*]YYYY
    {
        'name': nm.XSD_GYEAR_MONTH,
        'datatype': datatypes.GregorianYearMonth10,
        'python_type': datatypes.GregorianYearMonth10,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.GregorianYearMonth10.fromstring,
    },  # [-][Y*]YYYY-MM
)

XSD_11_BUILTIN_TYPES: tuple[dict[str, Any], ...] = XSD_COMMON_BUILTIN_TYPES + (
    {
        'name': nm.XSD_DOUBLE,
        'datatype': datatypes.DoubleProxy,
        'python_type': float,
        'admitted_facets': FLOAT_FACETS,
        'facets': [XSD11_FLOAT_PATTERN_ELEMENT, COLLAPSE_WHITE_SPACE_ELEMENT],
        'from_python': python_to_float,
    },  # 64 bit floating point
    {
        'name': nm.XSD_FLOAT,
        'datatype': datatypes.Float,
        'python_type': float,
        'admitted_facets': FLOAT_FACETS,
        'facets': [XSD11_FLOAT_PATTERN_ELEMENT, COLLAPSE_WHITE_SPACE_ELEMENT],
        'from_python': python_to_float,
    },  # 32 bit floating point

    # --- Year related primitive types (year 0 allowed and mapped to 1 BCE) ---
    {
        'name': nm.XSD_DATETIME,
        'datatype': datatypes.DateTime,
        'python_type': datatypes.DateTime,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.DateTime.fromstring,
    },  # [-][Y*]YYYY-MM-DD[Thh:mm:ss]
    {
        'name': nm.XSD_DATE,
        'datatype': datatypes.Date,
        'python_type': datatypes.Date,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.Date.fromstring,
    },  # [-][Y*]YYYY-MM-DD
    {
        'name': nm.XSD_GYEAR,
        'datatype': datatypes.GregorianYear,
        'python_type': datatypes.GregorianYear,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.GregorianYear.fromstring,
    },  # [-][Y*]YYYY
    {
        'name': nm.XSD_GYEAR_MONTH,
        'datatype': datatypes.GregorianYearMonth,
        'python_type': datatypes.GregorianYearMonth,
        'admitted_facets': DATETIME_FACETS,
        'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
        'to_python': datatypes.GregorianYearMonth.fromstring,
    },  # [-][Y*]YYYY-MM
    # --- Datetime derived types (XSD 1.1) ---
    {
        'name': nm.XSD_DATE_TIME_STAMP,
        'datatype': datatypes.DateTimeStamp,
        'python_type': datatypes.DateTimeStamp,
        'base_type': nm.XSD_DATETIME,
        'to_python': datatypes.DateTime.fromstring,
        'facets': [Element(nm.XSD_EXPLICIT_TIMEZONE, value='required')],
    },  # [-][Y*]YYYY-MM-DD[Thh:mm:ss] with required timezone
    {
        'name': nm.XSD_DAY_TIME_DURATION,
        'datatype': datatypes.DayTimeDuration,
        'python_type': datatypes.DayTimeDuration,
        'base_type': nm.XSD_DURATION,
        'to_python': datatypes.DayTimeDuration.fromstring,
    },  # PnYnMnDTnHnMnS with month a year equal to 0
    {
        'name': nm.XSD_YEAR_MONTH_DURATION,
        'datatype': datatypes.YearMonthDuration,
        'python_type': datatypes.YearMonthDuration,
        'base_type': nm.XSD_DURATION,
        'to_python': datatypes.YearMonthDuration.fromstring,
    },  # PnYnMnDTnHnMnS with day and time equals to 0
    # --- xs:error primitive type (XSD 1.1) ---
    {
        'name': nm.XSD_ERROR,
        'datatype': type(None),
        'python_type': type(None),
        'admitted_facets': (),
        'facets': [error_type_validator],
    },  # xs:error has no value space and no lexical space
)

BUILTIN_TYPES = {
    '1.0': XSD_10_BUILTIN_TYPES,
    '1.1': XSD_11_BUILTIN_TYPES
}



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
This module contains classes for XML Schema model groups.
"""
import warnings
from collections.abc import Iterable, Iterator, MutableMapping, MutableSequence
from copy import copy as _copy
from operator import attrgetter
from typing import TYPE_CHECKING, cast, overload, Any, Optional, Union
from xml.etree import ElementTree

import xmlschema.names as nm
from xmlschema.exceptions import XMLSchemaValueError
from xmlschema.aliases import ElementType, NsmapType, SchemaType, ModelParticleType, \
    SchemaElementType, ComponentClassType, OccursCounterType
from xmlschema.converters import ElementData
from xmlschema.translation import gettext as _
from xmlschema.utils.decoding import Empty, raw_encode_value, raw_encode_attributes
from xmlschema.utils.qnames import get_qname, local_name
from xmlschema import _limits
from xmlschema.caching import schema_cache, schema_cached_property

from .exceptions import XMLSchemaModelError, XMLSchemaModelDepthError, \
    XMLSchemaValidationError, XMLSchemaTypeTableWarning, XMLSchemaCircularityError, \
    XMLSchemaValidatorError
from .validation import ValidationContext, DecodeContext, EncodeContext, ValidationMixin
from .xsdbase import XsdComponent, XsdType
from .particles import ParticleMixin
from .elements import XsdElement, XsdAlternative
from .wildcards import XsdAnyElement, XsdOpenContent
from .models import ModelVisitor, InterleavedModelVisitor, SuffixedModelVisitor, \
    iter_unordered_content, iter_collapsed_content

if TYPE_CHECKING:
    from .complex_types import XsdComplexType  # noqa: F401
    from .particles import OccursCalculator  # noqa: F401

get_occurs = attrgetter('min_occurs', 'max_occurs')

ANY_ELEMENT = ElementTree.Element(
    nm.XSD_ANY,
    attrib={
        'namespace': '##any',
        'processContents': 'lax',
        'minOccurs': '0',
        'maxOccurs': 'unbounded'
    })

GroupDecodeType = Optional[list[tuple[Union[str, int], Any, Optional[SchemaElementType]]]]
GroupEncodeType = ElementType


class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
               ParticleMixin, ValidationMixin[ElementType, GroupDecodeType]):
    """
    Class for XSD 1.0 *model group* definitions.

    ..  <group
          id = ID
          maxOccurs = (nonNegativeInteger | unbounded) : 1
          minOccurs = nonNegativeInteger : 1
          name = NCName
          ref = QName
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, (all | choice | sequence)?)
        </group>

    ..  <all
          id = ID
          maxOccurs = 1 : 1
          minOccurs = (0 | 1) : 1
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, element*)
        </all>

    ..  <choice
          id = ID
          maxOccurs = (nonNegativeInteger | unbounded)  : 1
          minOccurs = nonNegativeInteger : 1
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, (element | group | choice | sequence | any)*)
        </choice>

    ..  <sequence
          id = ID
          maxOccurs = (nonNegativeInteger | unbounded)  : 1
          minOccurs = nonNegativeInteger : 1
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, (element | group | choice | sequence | any)*)
        </sequence>
    """
    parent: Optional[Union['XsdComplexType', 'XsdGroup']]
    model: str
    mixed: bool = False
    ref: Optional['XsdGroup']
    content: list[ModelParticleType]  # The effective content model for validate
    restriction: Optional['XsdGroup'] = None
    redefine: Optional['XsdGroup']

    # For XSD 1.1 openContent processing
    open_content: Optional[XsdOpenContent] = None
    _ADMITTED_TAGS = (nm.XSD_GROUP, nm.XSD_SEQUENCE, nm.XSD_ALL, nm.XSD_CHOICE)

    __slots__ = ('_group', 'content', 'oid', 'model', 'min_occurs', 'max_occurs')

    def __init__(self, elem: ElementType,
                 schema: SchemaType,
                 parent: Optional[Union['XsdComplexType', 'XsdGroup']] = None) -> None:

        self._group: list[ModelParticleType] = []
        self.content = self._group
        self.oid = (self,)
        self.min_occurs = self.max_occurs = 1
        super().__init__(elem, schema, parent)

    def __repr__(self) -> str:
        model = getattr(self, 'model', None)
        if self.name is None:
            return '%s(model=%r, occurs=%r)' % (
                self.__class__.__name__, model, list(self.occurs)
            )
        elif self.ref is None:
            return '%s(name=%r, model=%r, occurs=%r)' % (
                self.__class__.__name__, self.prefixed_name, model, list(self.occurs)
            )
        else:
            return '%s(ref=%r, model=%r, occurs=%r)' % (
                self.__class__.__name__, self.prefixed_name, model, list(self.occurs)
            )

    @overload
    def __getitem__(self, i: int) -> ModelParticleType: ...

    @overload
    def __getitem__(self, s: slice) -> MutableSequence[ModelParticleType]: ...

    def __getitem__(self, i: Union[int, slice]) \
            -> Union[ModelParticleType, MutableSequence[ModelParticleType]]:
        return self._group[i]

    def __setitem__(self, i: Union[int, slice], o: Any) -> None:
        if self._built:
            raise XMLSchemaValidatorError(self, 'cannot modify an already built group')
        self._group[i] = o

    def __delitem__(self, i: Union[int, slice]) -> None:
        if self._built:
            raise XMLSchemaValidatorError(self, 'cannot modify an already built group')
        del self._group[i]

    def __len__(self) -> int:
        return len(self._group)

    def insert(self, i: int, item: ModelParticleType) -> None:
        if self._built:
            raise XMLSchemaValidatorError(self, 'cannot modify an already built group')
        self._group.insert(i, item)

    def is_emptiable(self) -> bool:
        if self.model == 'choice':
            return self.min_occurs == 0 or not self or any(item.is_emptiable() for item in self)
        else:
            return self.min_occurs == 0 or not self or all(item.is_emptiable() for item in self)

    def is_single(self) -> bool:
        if self.max_occurs != 1 or not self:
            return False
        elif len(self) > 1 or not isinstance(self[0], XsdGroup):
            return True
        else:
            return self[0].is_single()

    def is_pointless(self, parent: 'XsdGroup') -> bool:
        """
        Returns `True` if the group may be eliminated without affecting the model,
        `False` otherwise. A group is pointless if one of those conditions is verified:

         - the group is empty
         - minOccurs == maxOccurs == 1 and the group has one child
         - minOccurs == maxOccurs == 1 and the group and its parent have a sequence model
         - minOccurs == maxOccurs == 1 and the group and its parent have a choice model

        Ref: https://www.w3.org/TR/2004/REC-xmlschema-1-20041028/#coss-particle

        :param parent: effective parent of the model group.
        """
        if not self:
            return True
        elif self.min_occurs != 1 or self.max_occurs != 1:
            return False
        elif len(self) == 1:
            return True
        elif self.model == 'sequence' and parent.model != 'sequence':
            return False
        elif self.model == 'choice' and parent.model != 'choice':
            return False
        else:
            return True

    @property
    def open_content_mode(self) -> str:
        return 'none' if self.open_content is None else self.open_content.mode

    @property
    def effective_min_occurs(self) -> int:
        if not self.min_occurs or not self:
            return 0

        effective_items: list[Any]
        min_occurs: int
        effective_items = [e for e in self.iter_model() if e.effective_max_occurs != 0]
        if not effective_items:
            return 0
        elif self.model == 'choice':
            min_occurs = min(e.effective_min_occurs for e in effective_items)
            return self.min_occurs * min_occurs
        elif self.model == 'all':
            min_occurs = max(e.effective_min_occurs for e in effective_items)
            return min_occurs

        not_emptiable_items = [e for e in effective_items if e.effective_min_occurs]
        if not not_emptiable_items:
            return 0
        elif len(not_emptiable_items) > 1:
            return self.min_occurs

        min_occurs = not_emptiable_items[0].effective_min_occurs
        return self.min_occurs * min_occurs

    @property
    def effective_max_occurs(self) -> Optional[int]:
        if self.max_occurs == 0 or not self:
            return 0

        effective_items: list[Any]
        max_occurs: int

        model_items = [(e, e.effective_max_occurs) for e in self.iter_model()]
        effective_items = [x for x in model_items if x[1] != 0]
        if not effective_items:
            return 0
        elif self.max_occurs is None:
            return None
        elif self.model == 'choice':
            if any(x[1] is None for x in effective_items):
                return None
            else:
                max_occurs = max(x[1] for x in effective_items)
                return self.max_occurs * max_occurs

        not_emptiable_items = [x for x in effective_items if x[0].effective_min_occurs]
        if not not_emptiable_items:
            if any(x[1] is None for x in effective_items):
                return None
            else:
                max_occurs = max(x[1] for x in effective_items)
                return self.max_occurs * max_occurs

        elif len(not_emptiable_items) > 1:
            if self.model == 'sequence':
                return self.max_occurs
            elif all(x[1] is None for x in not_emptiable_items):
                return None
            else:
                max_occurs = min(x[1] for x in not_emptiable_items if x[1] is not None)
                return max_occurs
        elif not_emptiable_items[0][1] is None:
            return None
        else:
            return self.max_occurs * cast(int, not_emptiable_items[0][1])

    def has_occurs_restriction(
            self, other: Union[ModelParticleType, ParticleMixin, 'OccursCalculator']) -> bool:

        if not self:
            return True
        elif isinstance(other, XsdGroup):
            return super().has_occurs_restriction(other)

        # Group particle compared to element particle
        if self.max_occurs is None or any(e.max_occurs is None for e in self):
            if other.max_occurs is not None:
                return False
            elif self.model == 'choice':
                return self.min_occurs * min(e.min_occurs for e in self) >= other.min_occurs
            else:
                return self.min_occurs * sum(e.min_occurs for e in self) >= other.min_occurs

        elif self.model == 'choice':
            if self.min_occurs * min(e.min_occurs for e in self) < other.min_occurs:
                return False
            elif other.max_occurs is None:
                return True
            else:
                value: int
                try:
                    value = max(e.max_occurs for e in self)  # type: ignore[type-var, assignment]
                except TypeError:
                    return False
                else:
                    return self.max_occurs * value <= other.max_occurs

        else:
            if self.min_occurs * sum(e.min_occurs for e in self) < other.min_occurs:
                return False
            elif other.max_occurs is None:
                return True
            else:
                try:
                    value = sum(e.max_occurs for e in self)  # type: ignore[misc]
                except TypeError:
                    return False
                else:
                    return self.max_occurs * value <= other.max_occurs

    def iter_model(self) -> Iterator[ModelParticleType]:
        """
        A generator function iterating elements and groups of a model group.
        Skips pointless groups, iterating deeper through them.
        Raises `XMLSchemaModelDepthError` if the *depth* of the model is over
        `limits.MAX_MODEL_DEPTH` value.
        """
        iterators: list[Iterator[ModelParticleType]] = []
        particles = iter(self)

        while True:
            for item in particles:
                if isinstance(item, XsdGroup) and item.is_pointless(parent=self):
                    iterators.append(particles)
                    particles = iter(item)
                    if len(iterators) > _limits.MAX_MODEL_DEPTH:
                        raise XMLSchemaModelDepthError(self)
                    break
                else:
                    yield item
            else:
                try:
                    particles = iterators.pop()
                except IndexError:
                    return

    def iter_elements(self) -> Iterator[SchemaElementType]:
        """
        A generator function iterating model's elements. Raises `XMLSchemaModelDepthError`
        if the overall depth of the model groups is over `limits.MAX_MODEL_DEPTH`.
        """
        if self.max_occurs == 0:
            return

        iterators: list[Iterator[ModelParticleType]] = []
        particles = iter(self)

        while True:
            for item in particles:
                if isinstance(item, XsdGroup):
                    if item.max_occurs == 0:
                        continue

                    iterators.append(particles)
                    particles = iter(item.content)
                    if len(iterators) > _limits.MAX_MODEL_DEPTH:
                        raise XMLSchemaModelDepthError(self)
                    break
                else:
                    yield item
            else:
                try:
                    particles = iterators.pop()
                except IndexError:
                    return

    @schema_cached_property
    def elements(self) -> tuple[SchemaElementType, ...]:
        return tuple(self.iter_elements())

    def get_subgroups(self, particle: ModelParticleType) -> list['XsdGroup']:
        """
        Returns a list of the groups that represent the first path to the enclosed particle.
        Raises an `XMLSchemaModelError` if the argument is not a particle of the model group.
        """
        for subgroups in self.iter_subgroups(particle):
            return subgroups
        else:
            return []

    def iter_subgroups(self, particle: ModelParticleType) -> Iterator[list['XsdGroup']]:
        """
        Iterates all subgroups where the provided particle is present.
        Raises an `XMLSchemaModelError` if the argument is not a particle of the model group.
        """
        subgroups: list[tuple[XsdGroup, Iterator[ModelParticleType]]] = []
        group, children = self, iter(self if self.ref is None else self.ref)
        found = False

        while True:
            for child in children:
                if child is particle:
                    found = True
                    _subgroups = [x[0] for x in subgroups]
                    _subgroups.append(group)
                    yield _subgroups
                elif isinstance(child, XsdGroup):
                    if len(subgroups) > _limits.MAX_MODEL_DEPTH:
                        raise XMLSchemaModelDepthError(self)
                    subgroups.append((group, children))
                    group, children = child, iter(child if child.ref is None else child.ref)
                    break
            else:
                try:
                    group, children = subgroups.pop()
                except IndexError:
                    if not found:
                        msg = _('{!r} is not a particle of the model group')
                        raise XMLSchemaModelError(self, msg.format(particle)) from None
                    return

    def get_model_visitor(self) -> ModelVisitor:
        if self.open_content is None or self.open_content.mode == 'none':
            return ModelVisitor(self)
        elif self.open_content.mode == 'interleave':
            return InterleavedModelVisitor(self, self.open_content.any_element)
        else:
            return SuffixedModelVisitor(self, self.open_content.any_element)

    def overall_min_occurs(self, particle: ModelParticleType) -> int:
        """
        Returns the overall min occurs of a particle in the model group.
        """
        model = self.get_model_visitor()
        return model.overall_min_occurs(particle)

    def overall_max_occurs(self, particle: ModelParticleType) -> Optional[int]:
        """
        Returns the overall max occurs of a particle in the model group.
        """
        model = self.get_model_visitor()
        return model.overall_max_occurs(particle)

    def is_optional(self, particle: ModelParticleType) -> bool:
        """
        Returns `True` if the provided particle can be optional in the model group.
        """
        return self.overall_min_occurs(particle) == 0

    def is_missing(self, occurs: OccursCounterType) -> bool:
        value = occurs[self.oid] or occurs[self]
        return not self.is_emptiable() if value == 0 else self.min_occurs > value

    def get_expected(self, occurs: OccursCounterType) -> list[SchemaElementType]:
        """
        Returns the expected elements of the current and descendant groups
        given a counter of occurrences. Returns an empty list if the group
        reached the maximum number of occurrences.
        """
        expected: list[SchemaElementType] = []

        if self.is_over(occurs):
            return expected

        for e in self.elements:
            if e not in expected and isinstance(e, XsdElement) and e.min_occurs > occurs[e]:
                expected.append(e)
                expected.extend(s for s in e.iter_substitutes())

        return expected

    def __copy__(self) -> 'XsdGroup':
        group: XsdGroup = object.__new__(self.__class__)
        group.__dict__.update(self.__dict__)

        for attr in self._mro_slots():
            object.__setattr__(group, attr, getattr(self, attr))

        group.errors = self.errors.copy()
        group._group = self._group.copy()
        if self.ref is None:
            group.content = group._group
        return group

    def _any_content_group_fallback(self) -> None:
        self.model = 'sequence'
        self.mixed = True
        xsd_element = self.builders.any_element_class(ANY_ELEMENT, self.schema, self)
        self._group.clear()
        self._group.append(xsd_element)
        self.__dict__.pop('elements', None)

    def _parse(self) -> None:
        self._group.clear()
        self.__dict__.pop('elements', None)
        self._parse_particle(self.elem)

        if self.parent is not None and self.parent.mixed:
            self.mixed = self.parent.mixed

        if self.elem.tag != nm.XSD_GROUP:
            # Local group (sequence|all|choice)
            if 'name' in self.elem.attrib:
                msg = _("attribute 'name' not allowed in a local group")
                self.parse_error(msg)
            self._parse_content_model(self.elem)

        elif self._parse_reference():
            assert self.name is not None
            try:
                xsd_group = self.maps.groups[self.name]
            except KeyError:
                self.parse_error(_("missing group %r") % self.prefixed_name)
                self._any_content_group_fallback()
            except XMLSchemaCircularityError as err:
                # Circular definition, substituted with any content group.
                self.parse_error(err, err.elem)
                self._any_content_group_fallback()
            else:
                self.model = xsd_group.model
                if self.model == 'all':
                    if self.max_occurs != 1:
                        msg = _("maxOccurs must be 1 for 'all' model groups")
                        self.parse_error(msg)
                    if self.min_occurs not in (0, 1):
                        msg = _("minOccurs must be (0 | 1) for 'all' model groups")
                        self.parse_error(msg)
                    if self.xsd_version == '1.0' and isinstance(self.parent, XsdGroup):
                        msg = _("in XSD 1.0 an 'all' model group cannot be nested")
                        self.parse_error(msg)
                self._group.append(xsd_group)
                self.ref = xsd_group
                self.target_namespace = xsd_group.target_namespace
                self.content = xsd_group.content

        else:
            attrib = self.elem.attrib
            try:
                self.name = get_qname(self.target_namespace, attrib['name'])
            except KeyError:
                pass
            else:
                if self.parent is not None:
                    msg = _("attribute 'name' not allowed in a local group")
                    self.parse_error(msg)
                else:
                    if 'minOccurs' in attrib:
                        msg = _("attribute 'minOccurs' not allowed in a global group")
                        self.parse_error(msg)
                    if 'maxOccurs' in attrib:
                        msg = _("attribute 'maxOccurs' not allowed in a global group")
                        self.parse_error(msg)

                content_model = self._parse_child_component(self.elem, strict=True)
                if content_model is not None:
                    if self.parent is None:
                        if 'minOccurs' in content_model.attrib:
                            msg = _("attribute 'minOccurs' not allowed in a global group")
                            self.parse_error(msg, content_model)
                        if 'maxOccurs' in content_model.attrib:
                            msg = _("attribute 'maxOccurs' not allowed in a global group")
                            self.parse_error(msg, content_model)

                    if content_model.tag in nm.MODEL_TAGS:
                        self._parse_content_model(content_model)
                    else:
                        msg = _('unexpected tag %r')
                        self.parse_error(msg % content_model.tag, content_model)
                        self._any_content_group_fallback()

    def _parse_content_model(self, content_model: ElementType) -> None:
        self.model = local_name(content_model.tag)
        if self.model == 'all':
            if self.max_occurs != 1:
                msg = _("maxOccurs must be 1 for 'all' model groups")
                self.parse_error(msg)
            if self.min_occurs not in (0, 1):
                msg = _("minOccurs must be (0 | 1) for 'all' model groups")
                self.parse_error(msg)

        child: ElementType
        for child in content_model:
            match child.tag:
                case x if callable(x):
                    continue
                case nm.XSD_ANNOTATION:
                    continue
                case nm.XSD_ELEMENT:
                    self._group.append(self.builders.element_class(child, self.schema, self))
                case nm.XSD_ALL:
                    self.parse_error(_("'all' model can contain only elements"))
                case nm.XSD_ANY:
                    self._group.append(
                        self.builders.any_element_class(child, self.schema, self)
                    )
                case nm.XSD_SEQUENCE | nm.XSD_CHOICE:
                    self._group.append(XsdGroup(child, self.schema, self))
                case nm.XSD_GROUP:
                    try:
                        ref = self.schema.resolve_qname(child.attrib['ref'])
                    except (KeyError, ValueError, RuntimeError) as err:
                        if 'ref' not in child.attrib:
                            msg = _("missing attribute 'ref' in local group")
                            self.parse_error(msg, child)
                        else:
                            self.parse_error(err, child)
                        continue

                    if ref != self.name:
                        xsd_group = XsdGroup(child, self.schema, self)
                        if xsd_group.model == 'all':
                            msg = _("'all' model can appears only at 1st level of a model group")
                            self.parse_error(msg)
                        else:
                            self._group.append(xsd_group)
                    elif self.redefine is not None:
                        self._group.append(self.redefine)
                        if child.get('minOccurs', '1') != '1' \
                                or child.get('maxOccurs', '1') != '1':
                            msg = _("Redefined group reference can't have "
                                    "minOccurs/maxOccurs other than 1")
                            self.parse_error(msg)
                    else:
                        msg = _("Circular definition detected for group %r")
                        self.parse_error(msg % self.name)

    def build(self) -> None:
        if self._built is False:
            self._built = None
            try:
                for item in self._group:
                    if isinstance(item, XsdElement):
                        item.build()

                if self.redefine is not None:
                    for group in self.redefine.iter_components(XsdGroup):
                        group.build()
                self._built = True
            finally:
                if self._built is None:
                    self._built = False

    @property
    def schema_elem(self) -> ElementType:
        return self.parent.elem if self.parent is not None else self.elem

    def iter_components(self, xsd_classes: Optional[ComponentClassType] = None) \
            -> Iterator[XsdComponent]:
        if xsd_classes is None or isinstance(self, xsd_classes):
            yield self
        for item in self:
            if item.parent is None:
                continue
            elif item.parent is not self.parent and isinstance(item.parent, XsdType) \
                    and item.parent.parent is None:
                continue
            yield from item.iter_components(xsd_classes)

        if self.redefine is not None and self.redefine not in self:
            yield from self.redefine.iter_components(xsd_classes)

    def admits_restriction(self, model: str) -> bool:
        if self.model == model:
            return True
        elif self.model == 'all':
            return model == 'sequence'
        elif self.model == 'choice':
            return model == 'sequence' or len(self.ref or self) <= 1
        else:
            return model == 'choice' or len(self.ref or self) <= 1

    def is_empty(self) -> bool:
        return not self.mixed and (not self._group or self.max_occurs == 0)

    @schema_cache
    def is_restriction(self, other: ModelParticleType, check_occurs: bool = True) -> bool:
        if not self._group:
            return True
        elif not isinstance(other, ParticleMixin):
            raise XMLSchemaValueError("the argument 'other' must be an XSD particle")
        elif not isinstance(other, XsdGroup):
            return self.is_element_restriction(other)
        elif not other:
            return False
        elif len(other) == other.min_occurs == other.max_occurs == 1:
            if len(self) > 1:
                return self.is_restriction(other[0], check_occurs)
            elif self.ref is None and isinstance(self[0], XsdGroup) \
                    and self[0].is_pointless(parent=self):
                return self[0].is_restriction(other[0], check_occurs)

        # Compare model with model
        if self.model != other.model and self.model != 'sequence' and \
                (len(self) > 1 or self.ref is not None and len(self[0]) > 1):
            return False
        elif self.model == other.model or other.model == 'sequence':
            return self.is_sequence_restriction(other)
        elif other.model == 'all':
            return self.is_all_restriction(other)
        else:  # other.model == 'choice':
            return self.is_choice_restriction(other)

    @schema_cache
    def is_element_restriction(self, other: ModelParticleType) -> bool:
        if self.xsd_version == '1.0' and isinstance(other, XsdElement) and \
                not other.ref and other.name not in self.maps.substitution_groups:
            return False
        elif not self.has_occurs_restriction(other):
            return False
        elif self.model == 'choice':
            if all(e.is_substitute(other) for e in self):
                return True
            return any(e.is_restriction(other, False) for e in self)
        else:
            min_occurs = 0
            max_occurs: Optional[int] = 0
            for item in self.iter_model():
                if isinstance(item, XsdGroup):
                    return False
                elif item.min_occurs == 0 or item.is_restriction(other, False):
                    min_occurs += item.min_occurs
                    if max_occurs is not None:
                        if item.max_occurs is None:
                            max_occurs = None
                        else:
                            max_occurs += item.max_occurs
                    continue
                return False

            if min_occurs < other.min_occurs:
                return False
            elif max_occurs is None:
                return other.max_occurs is None
            elif other.max_occurs is None:
                return True
            else:
                return max_occurs <= other.max_occurs

    @schema_cache
    def is_sequence_restriction(self, other: 'XsdGroup') -> bool:
        if not self.has_occurs_restriction(other):
            return False

        check_occurs = other.max_occurs != 0

        # Same model: declarations must simply preserve order
        other_iterator = iter(other.iter_model())
        for item in self.iter_model():
            for other_item in other_iterator:
                if other_item is item or item.is_restriction(other_item, check_occurs):
                    break
                elif other.model == 'choice':
                    if item.max_occurs != 0:
                        continue
                    elif not other_item.is_matching(item.name):
                        continue
                    elif all(e.max_occurs == 0 for e in self.iter_model()):
                        return False
                    else:
                        break
                elif not other_item.is_emptiable():
                    return False
            else:
                return False

        if other.model != 'choice':
            for other_item in other_iterator:
                if not other_item.is_emptiable():
                    return False
        return True

    @schema_cache
    def is_all_restriction(self, other: 'XsdGroup') -> bool:
        if not self.has_occurs_restriction(other):
            return False

        check_occurs = other.max_occurs != 0
        if self.ref is None:
            restriction_items = [x for x in self]
        else:
            restriction_items = [x for x in self[0]]

        for other_item in other.iter_model():
            for item in restriction_items:
                if other_item is item or item.is_restriction(other_item, check_occurs):
                    break
            else:
                if not other_item.is_emptiable():
                    return False
                continue
            restriction_items.remove(item)

        return not bool(restriction_items)

    def is_choice_restriction(self, other: 'XsdGroup') -> bool:
        if self.ref is None:
            if self.parent is None and other.parent is not None:
                return False  # not allowed restriction in XSD 1.0
            restriction_items = [x for x in self]
        elif other.parent is None:
            restriction_items = [x for x in self[0]]
        else:
            return False  # not allowed restriction in XSD 1.0

        check_occurs = other.max_occurs != 0
        max_occurs: Optional[int] = 0
        other_max_occurs: Optional[int] = 0

        for other_item in other.iter_model():
            for item in restriction_items:
                if other_item is item or item.is_restriction(other_item, check_occurs):
                    if max_occurs is not None:
                        if item.max_occurs is None:
                            max_occurs = None
                        else:
                            max_occurs += item.max_occurs

                    if other_max_occurs is not None:
                        if other_item.max_occurs is None:
                            other_max_occurs = None
                        else:
                            other_max_occurs = max(other_max_occurs, other_item.max_occurs)
                    break
            else:
                continue
            restriction_items.remove(item)

        if restriction_items:
            return False
        elif other_max_occurs is None:
            if other.max_occurs != 0:
                return True
            other_max_occurs = 0
        elif other.max_occurs is None:
            if other_max_occurs != 0:
                return True
            other_max_occurs = 0
        else:
            other_max_occurs *= other.max_occurs

        if max_occurs is None:
            return self.max_occurs == 0
        elif self.max_occurs is None:
            return max_occurs == 0
        else:
            return other_max_occurs >= max_occurs * self.max_occurs

    def check_dynamic_context(self, elem: ElementType,
                              xsd_element: SchemaElementType,
                              model_element: SchemaElementType,
                              namespaces: NsmapType) -> None:

        if model_element is not xsd_element and isinstance(model_element, XsdElement):
            if 'substitution' in model_element.block \
                    or xsd_element.type and xsd_element.type.is_blocked(model_element):
                reason = _("substitution of %r is blocked") % model_element
                raise XMLSchemaValidationError(model_element, elem, reason)

        alternatives: Union[tuple[()], list[XsdAlternative]] = []
        if isinstance(xsd_element, XsdAnyElement):
            if xsd_element.process_contents == 'skip':
                return

            try:
                xsd_element = self.maps.elements[elem.tag]
            except KeyError:
                if self.schema.meta_schema is None:
                    # Meta-schema groups ignore xsi:type (issue #350)
                    return

                try:
                    type_name = elem.attrib[nm.XSI_TYPE].strip()
                except KeyError:
                    return
                else:
                    xsd_type = self.maps.get_instance_type(
                        type_name, self.maps.any_type, namespaces
                    )
            else:
                alternatives = xsd_element.alternatives
                try:
                    type_name = elem.attrib[nm.XSI_TYPE].strip()
                except KeyError:
                    xsd_type = xsd_element.type
                else:
                    xsd_type = self.maps.get_instance_type(
                        type_name, xsd_element.type, namespaces
                    )

        else:
            if nm.XSI_TYPE not in elem.attrib or self.schema.meta_schema is None:
                xsd_type = xsd_element.type
            else:
                alternatives = xsd_element.alternatives
                try:
                    type_name = elem.attrib[nm.XSI_TYPE].strip()
                except KeyError:
                    xsd_type = xsd_element.type
                else:
                    xsd_type = self.maps.get_instance_type(
                        type_name, xsd_element.type, namespaces
                    )

            if model_element is not xsd_element and \
                    isinstance(model_element, XsdElement) and model_element.block:
                for derivation in model_element.block.split():
                    if xsd_type is not model_element.type and \
                            xsd_type.is_derived(model_element.type, derivation):
                        reason = _("usage of {0!r} with type {1} is blocked by "
                                   "head element").format(xsd_element, derivation)
                        raise XMLSchemaValidationError(self, elem, reason)

            if nm.XSI_TYPE not in elem.attrib or self.schema.meta_schema is None:
                return

        # If it's a restriction the context is the base_type's group
        group = self.restriction if self.restriction is not None else self

        # Dynamic EDC check of matched element
        for e in group.elements:
            if not isinstance(e, XsdElement):
                continue
            elif (other := e.match(elem.tag)) is None:
                continue

            if len(other.alternatives) != len(alternatives) or \
                    not xsd_type.is_dynamic_consistent(other.type):
                reason = _("{0!r} that matches {1!r} is not consistent with local "
                           "declaration {2!r}").format(elem, xsd_element, other)
                raise XMLSchemaValidationError(self, reason)

            if not all(any(a == x for x in alternatives) for a in other.alternatives) or \
                    not all(any(a == x for x in other.alternatives) for a in alternatives):
                msg = _("Maybe a not equivalent type table between elements "
                        "{0!r} and {1!r}.").format(self, xsd_element)
                warnings.warn(msg, XMLSchemaTypeTableWarning, stacklevel=3)

    @schema_cache
    def match_element(self, name: str) -> Optional[SchemaElementType]:
        """
        Try a model-less match of a child element. Returns the
        matched element, or `None` if there is no match.
        """
        for xsd_element in self.elements:
            if xsd_element.is_matching(name, group=self):
                return xsd_element
        return None

    def raw_decode(self, obj: ElementType, validation: str, context: ValidationContext) \
            -> GroupDecodeType:
        """
        Decoding an Element content.

        :param obj: an Element.
        :param validation: the validation mode. Can be 'lax', 'strict' or 'skip.
        :param context: the encoding context.
        :return: a list of 3-tuples (key, decoded data, decoder).
        """
        result: GroupDecodeType = None if context.validation_only else []
        cdata_index = 1  # keys for CDATA sections are positive integers
        index = 0

        if not self._group and self.model == 'choice' and self.min_occurs:
            reason = _("an empty 'choice' group with minOccurs > 0 cannot validate any content")
            context.validation_error(validation, self, reason, obj)
            return result

        if not self.mixed:
            # Check element CDATA
            if obj.text and obj.text.strip() or \
                    any(child.tail and child.tail.strip() for child in obj):
                if len(self) == 1 and isinstance(self[0], XsdAnyElement):
                    pass  # [XsdAnyElement()] equals to an empty complexType declaration
                else:
                    reason = _("character data between child elements not allowed")
                    context.validation_error(validation, self, reason, obj)
                    cdata_index = 0  # Do not decode CDATA

        if cdata_index and obj.text is not None:
            if self.mixed and context.preserve_mixed:
                text = obj.text
            else:
                text = str(obj.text.strip())
            if text:
                if result is not None:
                    result.append((cdata_index, text, None))
                cdata_index += 1

        over_max_depth = context.max_depth is not None and context.max_depth <= context.level

        errors: list[tuple[int, ModelParticleType, int, Optional[list[SchemaElementType]]]]
        xsd_element: Optional[SchemaElementType]
        expected: Optional[list[SchemaElementType]]

        errors = []
        broken_model = False
        namespaces = context.namespaces
        model = self.get_model_visitor()

        for index, child in enumerate(obj):
            if callable(child.tag):
                continue  # child is a comment or PI

            context.converter.set_xmlns_context(child, context.level)
            name = context.converter.map_qname(child.tag)

            while model.element is not None:
                xsd_element = model.match_element(child.tag)
                if xsd_element is None:
                    for particle, occurs, expected in model.advance(False):
                        errors.append((index, particle, occurs, expected))
                        model.clear()
                        broken_model = True  # the model is broken, continues with raw decoding.
                        xsd_element = self.match_element(child.tag)
                        break
                    else:
                        continue
                    break

                try:
                    self.check_dynamic_context(child, xsd_element, model.element, namespaces)
                except (XMLSchemaValidationError, TypeError) as err:
                    context.validation_error(validation, self, err, obj)

                for particle, occurs, expected in model.advance(True):
                    errors.append((index, particle, occurs, expected))
                break
            else:
                xsd_element = self.match_element(child.tag)
                if xsd_element is None:
                    errors.append((index, self, 0, None))
                    broken_model = True
                elif not broken_model:
                    errors.append((index, xsd_element, 0, []))
                    broken_model = True

            # Optional checks on matched XSD child
            if not isinstance(context, DecodeContext):
                if xsd_element is None or over_max_depth:
                    continue
            elif xsd_element is None:
                if context.keep_unknown:
                    result_item = self.maps.any_type.raw_decode(child, validation, context)
                    if result is not None:
                        result.append((name, result_item, None))
                continue
            elif over_max_depth:
                if context.depth_filler is not None and isinstance(xsd_element, XsdElement):
                    func = context.depth_filler
                    if result is not None:
                        result.append((name, func(xsd_element), xsd_element))
                continue

            result_item = xsd_element.raw_decode(child, validation, context)
            if result_item is Empty:
                continue
            elif result is not None:
                result.append((name, result_item, xsd_element))

                if cdata_index and child.tail is not None:
                    if self.mixed and context.preserve_mixed:
                        tail = child.tail
                    else:
                        tail = str(child.tail.strip())
                    if tail:
                        if result and isinstance(result[-1][0], int):
                            tail = result[-1][1] + ' ' + tail
                            result[-1] = result[-1][0], tail, None
                        else:
                            result.append((cdata_index, tail, None))
                            cdata_index += 1

        if model.element is not None:
            index = len(obj)
            for particle, occurs, expected in model.stop():
                errors.append((index, particle, occurs, expected))
                break

        if errors:
            for index, particle, occurs, expected in errors:
                context.children_validation_error(
                    validation, self, obj, index, particle, occurs, expected
                )

        if index or result is not None or obj.text is None:
            return result
        elif self.mixed and context.preserve_mixed:
            return [(1, obj.text, None)]
        else:
            return [(1, str(obj.text.strip()), None)]

    def raw_encode(self, obj: ElementData, validation: str, context: EncodeContext) \
            -> GroupEncodeType:
        """
        Encode data to a list containing Element children.

        :param obj: an ElementData instance.
        :param validation: the validation mode. Can be 'lax', 'strict' or 'skip'.
        :param context: the encoding context.
        :return: returns a couple with the text of the Element and a list of child \
        elements.
        """
        errors = []
        text = raw_encode_value(obj.text)
        children: list[ElementType] = []
        default_namespace = context.converter.get('')

        if (elem := context.elem) is None:
            context.elem = elem = context.create_element(tag=obj.tag)
            elem.attrib.update(raw_encode_attributes(obj.attributes))

        index = cdata_index = 0
        wrong_content_type = False
        over_max_depth = context.max_depth is not None and context.max_depth <= context.level
        model = self.get_model_visitor()

        content: Iterable[Any]
        if not obj.content:
            content = []
        elif isinstance(obj.content, MutableMapping) or context.unordered:
            content = iter_unordered_content(obj.content, self)
        elif not isinstance(obj.content, MutableSequence):
            wrong_content_type = True
            content = []
        elif not isinstance(obj.content[0], tuple):
            if len(obj.content) > 1 or text is not None:
                wrong_content_type = True
            else:
                text = raw_encode_value(obj.content[0])
            content = []
        elif context.converter.losslessly:
            content = obj.content
        else:
            content = iter_collapsed_content(obj.content, self)

        for index, (name, value) in enumerate(content):
            if isinstance(name, int):
                if not children:
                    text = text + value if text is not None else value
                else:
                    children[-1].tail = value
                cdata_index += 1
                continue

            xsd_element: Optional[SchemaElementType]
            while model.element is not None:
                xsd_element = model.match_element(name)
                if xsd_element is None:
                    for particle, occurs, expected in model.advance():
                        errors.append((index - cdata_index, particle, occurs, expected))
                    continue
                elif isinstance(xsd_element, XsdAnyElement):
                    value = get_qname(default_namespace, name), value

                for particle, occurs, expected in model.advance(True):
                    errors.append((index - cdata_index, particle, occurs, expected))
                break
            else:
                errors.append((index - cdata_index, self, 0, []))
                xsd_element = self.match_element(name)
                if isinstance(xsd_element, XsdAnyElement):
                    value = get_qname(default_namespace, name), value
                elif xsd_element is None:
                    if name.startswith('{') or ':' not in name:
                        reason = _('{!r} does not match any declared element '
                                   'of the model group').format(name)
                    else:
                        reason = _('{0} has an unknown prefix {1!r}').format(
                            name, name.split(':')[0]
                        )
                    context.validation_error(validation, self, reason, value)
                    continue

            if xsd_element.skip and not context.process_skipped:
                continue
            if over_max_depth:
                continue

            child = xsd_element.raw_encode(value, validation, context)
            if children is not None and child is not None:
                children.append(child)

        if model.element is not None:
            for particle, occurs, expected in model.stop():
                errors.append((index - cdata_index + 1, particle, occurs, expected))
                break

        context.set_element_content(
            elem=elem,
            text=text,
            children=children,
            level=context.level,
            mixed=self.mixed and context.preserve_mixed
        )

        if wrong_content_type:
            reason = _("wrong content type {!r}").format(type(obj.content))
            context.validation_error(validation, self, reason, elem)

        if not self.mixed and text and text.strip() and self and \
                (len(self) > 1 or not isinstance(self[0], XsdAnyElement)):
            reason = _("character data between child elements not allowed")
            context.validation_error(validation, self, reason, elem)

        for index, particle, occurs, expected in errors:
            context.children_validation_error(
                validation, self, elem, index, particle, occurs, expected
            )

        return elem


class Xsd11Group(XsdGroup):
    """
    Class for XSD 1.1 *model group* definitions.

    .. The XSD 1.1 model groups differ from XSD 1.0 groups for the 'all' model,
       that can contains also other groups.
    ..  <all
          id = ID
          maxOccurs = (0 | 1) : 1
          minOccurs = (0 | 1) : 1
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, (element | any | group)*)
        </all>
    """
    def _parse_content_model(self, content_model: ElementType) -> None:
        self.model = local_name(content_model.tag)
        if self.model == 'all':
            if self.max_occurs not in (0, 1):
                msg = _("maxOccurs must be (0 | 1) for 'all' model groups")
                self.parse_error(msg)
            if self.min_occurs not in (0, 1):
                msg = _("minOccurs must be (0 | 1) for 'all' model groups")
                self.parse_error(msg)

        for child in content_model:
            match child.tag:
                case nm.XSD_ELEMENT:
                    self._group.append(self.builders.element_class(child, self.schema, self))
                case nm.XSD_ANY:
                    self._group.append(self.builders.any_element_class(child, self.schema, self))
                case x if x in nm.MODEL_TAGS:
                    self._group.append(Xsd11Group(child, self.schema, self))
                case nm.XSD_GROUP:
                    try:
                        ref = self.schema.resolve_qname(child.attrib['ref'])
                    except (KeyError, ValueError, RuntimeError) as err:
                        if 'ref' not in child.attrib:
                            msg = _("missing attribute 'ref' in local group")
                            self.parse_error(msg, child)
                        else:
                            self.parse_error(err, child)
                        continue

                    if ref != self.name:
                        xsd_group = Xsd11Group(child, self.schema, self)
                        self._group.append(xsd_group)
                        if (self.model != 'all') ^ (xsd_group.model != 'all'):
                            msg = _("an xs:{0} group cannot include a reference to an "
                                    "xs:{1} group").format(self.model, xsd_group.model)
                            self.parse_error(msg)
                            self._group.pop()

                    elif self.redefine is not None:
                        if child.get('minOccurs', '1') != '1' or child.get('maxOccurs', '1') != '1':
                            msg = _("Redefined group reference cannot have "
                                    "minOccurs/maxOccurs other than 1")
                            self.parse_error(msg)
                        self._group.append(self.redefine)
                    else:
                        msg = _("Circular definition detected for group %r")
                        self.parse_error(msg % self.name)

    def admits_restriction(self, model: str) -> bool:
        if self.model == model or self.model == 'all':
            return True
        elif self.model == 'choice':
            return model == 'sequence' or len(self.ref or self) <= 1
        else:
            return model == 'choice' or len(self.ref or self) <= 1

    def is_restriction(self, other: ModelParticleType, check_occurs: bool = True) -> bool:
        if not self._group:
            return True
        elif not isinstance(other, ParticleMixin):
            raise XMLSchemaValueError("the argument 'base' must be a %r instance" % ParticleMixin)
        elif not isinstance(other, XsdGroup):
            return self.is_element_restriction(other)
        elif not other:
            return False
        elif len(other) == other.min_occurs == other.max_occurs == 1:
            if len(self) > 1:
                return self.is_restriction(other[0], check_occurs)
            elif self.ref is None and isinstance(self[0], XsdGroup) \
                    and self[0].is_pointless(parent=self):
                return self[0].is_restriction(other[0], check_occurs)

        if other.model == 'sequence':
            return self.is_sequence_restriction(other)
        elif other.model == 'all':
            return self.is_all_restriction(other)
        else:  # other.model == 'choice':
            return self.is_choice_restriction(other)

    def has_occurs_restriction(
            self, other: Union[ModelParticleType, ParticleMixin, 'OccursCalculator']) -> bool:
        if not isinstance(other, XsdGroup):
            return super().has_occurs_restriction(other)
        elif not self:
            return True
        elif self.effective_min_occurs < other.effective_min_occurs:
            return False

        effective_max_occurs = self.effective_max_occurs
        if effective_max_occurs == 0:
            return True
        elif effective_max_occurs is None:
            return other.effective_max_occurs is None

        try:
            return effective_max_occurs <= other.effective_max_occurs  # type: ignore[operator]
        except TypeError:
            return True

    def is_sequence_restriction(self, other: XsdGroup) -> bool:
        if not self.has_occurs_restriction(other):
            return False

        check_occurs = other.max_occurs != 0

        item_iterator = iter(self.iter_model())
        item = next(item_iterator, None)

        for other_item in other.iter_model():
            if item is not None and item.is_restriction(other_item, check_occurs):
                item = next(item_iterator, None)
            elif not other_item.is_emptiable():
                break
        else:
            if item is None:
                return True

        # Restriction check failed: try another check without removing pointless groups
        item_iterator = iter(self)
        item = next(item_iterator, None)

        for other_item in other.iter_model():
            if item is not None and item.is_restriction(other_item, check_occurs):
                item = next(item_iterator, None)
            elif not other_item.is_emptiable():
                break
        else:
            if item is None:
                return True

        # Restriction check failed again: try checking other items against self
        other_items = other.iter_model()
        for other_item in other_items:
            if self.is_restriction(other_item, check_occurs):
                return all(x.is_emptiable() for x in other_items)
            elif not other_item.is_emptiable():
                return False
        else:
            return False

    def is_all_restriction(self, other: XsdGroup) -> bool:
        restriction_items = [x for x in self.iter_model()]

        base_items = [x for x in other.iter_model()]

        # If the base includes more wildcard, calculates and appends a
        # wildcard union for validating wildcard unions in restriction
        wildcards: list[XsdAnyElement] = []
        extended: list[XsdAnyElement] = []
        for w1 in base_items:
            if isinstance(w1, XsdAnyElement):
                for w2 in wildcards:
                    if w1.process_contents == w2.process_contents and \
                            get_occurs(w1) == get_occurs(w2):
                        w2.union(w1)
                        extended.append(w2)
                        break
                else:
                    wildcards.append(_copy(w1))

        base_items.extend(extended)

        if self.model != 'choice':
            restriction_wildcards = [e for e in restriction_items if isinstance(e, XsdAnyElement)]

            for other_item in base_items:
                min_occurs, max_occurs = 0, other_item.max_occurs
                for k in range(len(restriction_items) - 1, -1, -1):
                    item = restriction_items[k]

                    if item.is_restriction(other_item, check_occurs=False):
                        if max_occurs is None:
                            min_occurs += item.min_occurs
                        elif item.max_occurs is None or max_occurs < item.max_occurs or \
                                min_occurs + item.min_occurs > max_occurs:
                            continue
                        else:
                            min_occurs += item.min_occurs
                            max_occurs -= item.max_occurs

                        restriction_items.remove(item)
                        if not min_occurs or max_occurs == 0:
                            break
                else:
                    if self.model == 'all' and restriction_wildcards:
                        if not isinstance(other_item, XsdGroup) and other_item.type \
                                and other_item.type.name != nm.XSD_ANY_TYPE:

                            for w in restriction_wildcards:
                                if w.is_matching(other_item.name, self.target_namespace):
                                    return False

                if min_occurs < other_item.min_occurs:
                    break
            else:
                if not restriction_items:
                    return True
            return False

        # Restriction with a choice model: this a more complex case
        # because the not emptiable elements of the base group have
        # to be included in each item of the choice group.
        not_emptiable_items = {x for x in base_items if x.min_occurs}

        for other_item in base_items:
            min_occurs, max_occurs = 0, other_item.max_occurs
            for k in range(len(restriction_items) - 1, -1, -1):
                item = restriction_items[k]

                if item.is_restriction(other_item, check_occurs=False):
                    if max_occurs is None:
                        min_occurs += item.min_occurs
                    elif item.max_occurs is None or max_occurs < item.max_occurs or \
                            min_occurs + item.min_occurs > max_occurs:
                        continue
                    else:
                        min_occurs += item.min_occurs
                        max_occurs -= item.max_occurs

                    if not_emptiable_items:
                        if len(not_emptiable_items) > 1:
                            continue
                        if other_item not in not_emptiable_items:
                            continue

                    restriction_items.remove(item)
                    if not min_occurs or max_occurs == 0:
                        break

            if min_occurs < other_item.min_occurs:
                break
        else:
            if not restriction_items:
                return True

        if any(not isinstance(x, XsdGroup) for x in restriction_items):
            return False

        # If the remaining items are groups try to verify if they are all
        # restrictions of the 'all' group and if each group contains all
        # not emptiable elements.
        for group in restriction_items:
            if not group.is_restriction(other):
                return False

            for item in not_emptiable_items:
                for e in group:
                    if e.name == item.name:
                        break
                else:
                    return False
        else:
            return True

    def is_choice_restriction(self, other: XsdGroup) -> bool:
        restriction_items = [x for x in self.iter_model()]
        has_not_empty_item = any(e.max_occurs != 0 for e in restriction_items)

        check_occurs = other.max_occurs != 0
        max_occurs: Optional[int] = 0
        other_max_occurs: Optional[int] = 0

        for other_item in other.iter_model():
            for item in restriction_items:
                if other_item is item or item.is_restriction(other_item, check_occurs):
                    if max_occurs is not None:
                        effective_max_occurs = item.effective_max_occurs
                        if effective_max_occurs is None:
                            max_occurs = None
                        elif self.model == 'choice':
                            max_occurs = max(max_occurs, effective_max_occurs)
                        else:
                            max_occurs += effective_max_occurs

                    if other_max_occurs is not None:
                        effective_max_occurs = other_item.effective_max_occurs
                        if effective_max_occurs is None:
                            other_max_occurs = None
                        else:
                            other_max_occurs = max(other_max_occurs, effective_max_occurs)
                    break
                elif item.max_occurs != 0:
                    continue
                elif not other_item.is_matching(item.name):
                    continue
                elif has_not_empty_item:
                    break
                else:
                    return False
            else:
                continue
            restriction_items.remove(item)

        if restriction_items:
            return False
        elif other_max_occurs is None:
            if other.max_occurs != 0:
                return True
            other_max_occurs = 0
        elif other.max_occurs is None:
            if other_max_occurs != 0:
                return True
            other_max_occurs = 0
        else:
            other_max_occurs *= other.max_occurs

        if max_occurs is None:
            return self.max_occurs == 0
        elif self.max_occurs is None:
            return max_occurs == 0
        else:
            return other_max_occurs >= max_occurs * self.max_occurs



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections.abc import MutableMapping, MutableSequence
from typing import TYPE_CHECKING, Any

from xmlschema.aliases import NsmapType, BaseXsdType, XmlnsType
from xmlschema.exceptions import XMLSchemaTypeError
from xmlschema.names import XSD_ANY_TYPE
from xmlschema.utils.qnames import local_name

from .base import ElementData, stackable, XMLSchemaConverter

if TYPE_CHECKING:
    from xmlschema.validators import XsdElement


class BadgerFishConverter(XMLSchemaConverter):
    """
    XML Schema based converter class for Badgerfish convention.

    ref: http://www.sklar.com/badgerfish/
    ref: https://badgerfish.ning.com/

    :param namespaces: Map from namespace prefixes to URI.
    :param dict_class: dictionary class to use for decoded data. Default is `dict`.
    :param list_class: list class to use for decoded data. Default is `list`.
    """
    __slots__ = ()

    def __init__(self, namespaces: NsmapType | None = None,
                 dict_class: type[dict[str, Any]] | None = None,
                 list_class: type[list[Any]] | None = None,
                 **kwargs: Any) -> None:
        kwargs.update(attr_prefix='@', text_key='$', cdata_prefix='$')
        super().__init__(namespaces, dict_class, list_class, **kwargs)

    @property
    def lossy(self) -> bool:
        return False

    def get_xmlns_from_data(self, obj: Any) -> XmlnsType:
        if not self._use_namespaces or not isinstance(obj, MutableMapping) or '@xmlns' not in obj:
            return None
        return [(k if k != '$' else '', v) for k, v in obj['@xmlns'].items()]

    @stackable
    def element_decode(self, data: ElementData, xsd_element: 'XsdElement',
                       xsd_type: BaseXsdType | None = None, level: int = 0) -> Any:
        xsd_type = xsd_type or xsd_element.type

        tag = self.map_qname(data.tag)
        result_dict = self.dict_class(t for t in self.map_attributes(data.attributes))

        xmlns = self.get_effective_xmlns(data.xmlns, level, xsd_element)
        if self._use_namespaces and xmlns:
            result_dict['@xmlns'] = self.dict_class((k or '$', v) for k, v in xmlns)

        xsd_group = xsd_type.model_group
        if xsd_group is None or not data.content:
            if data.text is not None:
                result_dict['$'] = data.text
        else:
            has_single_group = xsd_group.is_single()
            for name, item, xsd_child in self.map_content(data.content):
                if name.startswith('$') and name[1:].isdigit():
                    result_dict[name] = item
                    continue

                assert isinstance(item, MutableMapping) and xsd_child is not None

                item = item[name]
                if name in result_dict:
                    other = result_dict[name]
                    if not isinstance(other, MutableSequence) or not other:
                        result_dict[name] = self.list_class((other, item))
                    elif isinstance(other[0], MutableSequence) or \
                            not isinstance(item, MutableSequence):
                        other.append(item)
                    else:
                        result_dict[name] = self.list_class((other, item))
                else:
                    if xsd_type.name == XSD_ANY_TYPE or \
                            has_single_group and xsd_child.is_single():
                        result_dict[name] = item
                    else:
                        result_dict[name] = self.list_class((item,))

        if self.dict_class is dict:
            return {tag: result_dict}
        return self.dict_class(((tag, result_dict),))

    @stackable
    def element_encode(self, obj: Any, xsd_element: 'XsdElement', level: int = 0) -> ElementData:
        if not isinstance(obj, MutableMapping):
            raise XMLSchemaTypeError(f"A dictionary expected, got {type(obj)} instead.")
        elif len(obj) != 1 or all(k.startswith(('$', '@')) for k in obj):
            tag = xsd_element.name
        else:
            key, value = next(iter(obj.items()))
            tag = self.unmap_qname(key, xmlns=self.get_xmlns_from_data(value))
            if xsd_element.is_matching(tag):
                obj = value
            elif not self.namespaces and local_name(tag) == xsd_element.local_name:
                obj = value
            else:
                tag = xsd_element.name

        text = None
        content: list[tuple[str | int, Any]] = []
        attributes = {}

        xmlns = self.set_xmlns_context(obj, level)

        for name, value in obj.items():
            if name == '@xmlns':
                continue
            elif name == '$':
                text = value
            elif name[0] == '$' and name[1:].isdigit():
                content.append((int(name[1:]), value))
            elif name[0] == '@':
                attr_name = name[1:]
                ns_name = self.unmap_qname(attr_name, xsd_element.attributes)
                attributes[ns_name] = value
            elif not isinstance(value, MutableSequence) or not value:
                ns_name = self.unmap_qname(name, xmlns=self.get_xmlns_from_data(value))
                content.append((ns_name, value))
            elif isinstance(value[0], (MutableMapping, MutableSequence)):
                ns_name = self.unmap_qname(name, xmlns=self.get_xmlns_from_data(value[0]))
                for item in value:
                    content.append((ns_name, item))
            else:
                ns_name = self.unmap_qname(name)
                xsd_child = xsd_element.match_child(ns_name)
                if xsd_child is not None:
                    if xsd_child.type and xsd_child.type.is_list():
                        content.append((ns_name, value))
                    else:
                        content.extend((ns_name, item) for item in value)
                else:
                    content.extend((ns_name, item) for item in value)

        return ElementData(tag, text, content, attributes, xmlns)



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from decimal import Decimal
from collections.abc import Iterator, MutableMapping, MutableSequence
from typing import Any, Optional, Union

from xmlschema.aliases import DecodedValueType, NumericValueType

DecodedAttributesType = Optional[MutableMapping[str, DecodedValueType]]


class EmptyType:
    _instance = None

    def __new__(cls) -> 'EmptyType':
        if cls._instance is None:
            cls._instance = super(EmptyType, cls).__new__(cls)
        return cls._instance


Empty = EmptyType()
"""A singleton instance for representing empty decode/encode results."""


def count_digits(number: NumericValueType) -> tuple[int, int]:
    """
    Counts the digits of a number.

    :param number: an int or a float or a Decimal or a string representing a number.
    :return: a couple with the number of digits of the integer part and \
    the number of digits of the decimal part.
    """
    if isinstance(number, str):
        number = str(Decimal(number)).lstrip('-+')
    elif isinstance(number, bytes):
        number = str(Decimal(number.decode())).lstrip('-+')
    else:
        number = str(number).lstrip('-+')

    if 'E' in number:
        significand, _, _exponent = number.partition('E')
    elif 'e' in number:
        significand, _, _exponent = number.partition('e')
    elif '.' not in number:
        return len(number.lstrip('0')), 0
    else:
        integer_part, _, decimal_part = number.partition('.')
        return len(integer_part.lstrip('0')), len(decimal_part.rstrip('0'))

    significand = significand.strip('0')
    exponent = int(_exponent)

    num_digits = len(significand) - 1 if '.' in significand else len(significand)
    if exponent > 0:
        return num_digits + exponent, 0
    else:
        return 0, num_digits - exponent - 1


def strictly_equal(obj1: object, obj2: object) -> bool:
    """Checks if the objects are equal and are of the same type."""
    return obj1 == obj2 and type(obj1) is type(obj2)


def raw_encode_value(value: DecodedValueType) -> Optional[str]:
    """Encodes a simple value to XML."""
    if isinstance(value, bool):
        return 'true' if value else 'false'
    elif isinstance(value, (list, tuple)):
        return ' '.join(e for e in (raw_encode_value(v) for v in value) if e is not None)
    elif isinstance(value, bytes):
        return value.decode()
    else:
        return str(value) if value is not None else None


def raw_encode_attributes(attributes: DecodedAttributesType = None) \
        -> dict[str, str]:
    attrib: dict[str, str] = {}
    if attributes:
        for k, v in attributes.items():
            value = raw_encode_value(v)
            if value is not None:
                attrib[k] = value
    return attrib


def iter_decoded_data(obj: Any, level: int = 0) \
        -> Iterator[tuple[Union[MutableMapping[Any, Any], MutableSequence[Any]], int]]:
    """
    Iterates a nested object composed by lists and dictionaries,
    pairing with the level depth.
    """
    if isinstance(obj, MutableMapping):
        yield obj, level
        for value in obj.values():
            yield from iter_decoded_data(value, level + 1)
    elif isinstance(obj, MutableSequence):
        yield obj, level
        for item in obj:
            yield from iter_decoded_data(item, level + 1)
